# -*- coding: utf-8 -*-
"""
Created on Thu Feb 25 15:42:47 2021

@author: u0135479
"""

import numpy as np
import warnings


def add_emergency_noise(tree_matrix, noise_level=1e-5):
    warnings.warn("MDS matrix has rank 0")
    noise = np.random.standard_normal(size=tree_matrix.shape)
    # now make noise for distance matrix symmetric:
    for i in range(tree_matrix.shape[0]):
        for j in range(i, tree_matrix.shape[1]):
            if j == i:
                noise[i, j] = 0
            else:
                noise[i, j] = noise[j, i]
    return tree_matrix + noise * noise_level


def tree_splits_to_vector(clf, idx1, split_weight=None):  # it's a SIMILARITY measure

    the_tree = clf.estimators_[idx1].tree_  # does it work for both binary and survival data??
    # store feature splits, excluding leafs which have none ( == -1)
    the_splits = the_tree.feature[the_tree.feature > -1]
    if split_weight == "by_samples":
        # store relative n_samples that go across each (internal) node
        the_weights = (the_tree.n_node_samples / the_tree.n_node_samples[0])[the_tree.feature > -1]
    elif split_weight == "simple":
        the_weights = None  # no weighting (all equal weights)
    else:
        raise KeyError(
            "split_weight_style = '{}' not recognized,\
                 accepted values are 'simple' and  'by_samples'."
        )
    tree_vector_no_pad = np.bincount(the_splits, weights=the_weights)
    # PROBLEM: some features might be missing (at the tail), fill with zeros
    # does this work for both binary and survival?
    tree_vector = np.zeros(clf.n_features_in_)
    # the bincount will ignore features with no splits that come after
    # the last feature with at least one split
    # we need to include them with extra 0-s to preserve vector length
    tree_vector[: len(tree_vector_no_pad)] += tree_vector_no_pad

    return tree_vector


def rule_splits_to_vector(clf, idx1, feature_represent, sample):  # it's a SIMILARITY measure

    the_tree = clf.estimators_[idx1].tree_  # does it work for both binary and survival data??

    if not isinstance(sample, np.ndarray):  # pd.Series or smth else ( list? or..?)
        sample = sample.to_numpy()

    # for t, idx in zip([idx1, idx2], [0,1]):
    tree_splits = clf[idx1].tree_.feature
    # surv_tree = clf[idx1]

    # keep only nodes ( indices) that go through sample's path
    tree_path = clf[idx1].tree_.decision_path(sample.reshape(1, -1).astype(np.float32))
    tree_path = tree_path.toarray().reshape(-1)

    # store feature splits along path ( exclude last element: it's a leaf!)
    path_splits = [tree_splits[i] for i in range(len(tree_splits)) if tree_path[i] > 0][:-1]

    if feature_represent in ["by_samples", "weighted"]:
        # store weights across all (internal) PATH nodes
        # assign weight proportional to n_samples going through
        the_weights = (the_tree.n_node_samples / the_tree.n_node_samples[0])[tree_path > 0][:-1]
    elif feature_represent == "simple":
        the_weights = None  # no weighting (all equal weights)
    else:
        raise KeyError(
            "split_weight_style = '{}' not recognized,\
                  accepted values are 'simple' or  'by_samples' and 'weighted' ."
        )

    rule_vec_no_pad = np.bincount(path_splits, weights=the_weights)

    # pad rule vector with zero's for tail features that have no splitting
    rule_vector = np.zeros(clf.n_features_in_)  # rule_representation
    rule_vector[: len(rule_vec_no_pad)] += rule_vec_no_pad

    return rule_vector



import pytest
import pandas as pd
import numpy as np
from bellatrex.datasets import (
    load_binary_data,
    load_regression_data,
    load_survival_data,
    load_mlc_data,
    load_mtr_data,
)

# Parametrize the test cases: (load_function, expected_y_shape, return_xy)
dataset_loaders = [
    (load_binary_data, (748,), True),         # Binary: single column y
    (load_regression_data, (1503,), True),     # Regression: single column y
    (load_survival_data, (500,), True),       # Survival: structured array
    (load_mlc_data, (193, 7), True),          # Multi-label: 7 target columns
    (load_mtr_data, (1137, 3), True),          # Multi-target: 3 target columns
]

@pytest.mark.parametrize("loader_func, expected_y_shape, return_xy", dataset_loaders)
def test_dataset_loader_output(loader_func, expected_y_shape, return_xy):
    X, y = loader_func(return_X_y=return_xy)

    assert isinstance(X, pd.DataFrame)
    assert len(X) == expected_y_shape[0]
    if isinstance(y, np.ndarray) and y.dtype.names is not None:
        # survival: structured array
        assert y.shape == expected_y_shape
        assert set(y.dtype.names) == {"event", "time"} or set(y.dtype.names) == set(y.dtype.fields)
    elif isinstance(y, pd.DataFrame):
        assert y.shape == expected_y_shape
    else:
        # y is a Series or 1D array
        assert len(y) == expected_y_shape[0]


@pytest.mark.parametrize("loader_func", [
    load_binary_data,
    load_regression_data,
    load_survival_data,
    load_mlc_data,
    load_mtr_data,
])
def test_dataset_loader_df(loader_func):
    df = loader_func(return_X_y=False)
    assert isinstance(df, pd.DataFrame)
    assert not df.empty



import pytest
import numpy as np
import matplotlib

matplotlib.use("Agg")  # Use non-interactive backend for tests
import matplotlib.pyplot as plt

from sklearn.datasets import load_iris
from sklearn.tree import DecisionTreeClassifier

try:
    from bellatrex.plot_tree_patch import _color_brew, plot_tree_patched, _MPLTreeExporter
except ImportError:
    from app.bellatrex.plot_tree_patch import _color_brew, plot_tree_patched, _MPLTreeExporter


def test_color_brew_length_and_type():
    n = 5
    colors = _color_brew(n)
    assert isinstance(colors, list)
    assert len(colors) == n
    for color in colors:
        assert isinstance(color, list)
        assert len(color) == 3
        assert all(isinstance(c, int) for c in color)
        assert all(0 <= c <= 255 for c in color)


def test_plot_tree_patched_runs():
    iris = load_iris()
    X, y = iris.data, iris.target
    clf = DecisionTreeClassifier(max_depth=2, random_state=0)
    clf.fit(X, y)
    fig, ax = plt.subplots()
    anns = plot_tree_patched(clf, ax=ax)
    assert isinstance(anns, list)
    assert all(hasattr(a, "get_window_extent") for a in anns)
    plt.close(fig)


def test_plot_tree_patched_with_feature_names():
    iris = load_iris()
    X, y = iris.data, iris.target
    clf = DecisionTreeClassifier(max_depth=1, random_state=0)
    clf.fit(X, y)
    fig, ax = plt.subplots()
    anns = plot_tree_patched(clf, feature_names=iris.feature_names, ax=ax)
    assert isinstance(anns, list)
    plt.close(fig)


def test_exporter_invalid_precision():
    with pytest.raises(ValueError):
        _MPLTreeExporter(precision=-1)
    with pytest.raises(ValueError):
        _MPLTreeExporter(precision="bad")


@pytest.mark.skipif(
    pytest.importorskip("sksurv", reason="sksurv not installed") is None,
    reason="sksurv not available",
)
def test_plot_tree_patched_survivaltree():
    from sksurv.tree import SurvivalTree

    X = np.random.rand(20, 2)
    y = np.array([(True, 1.0 + i) for i in range(20)], dtype=[("event", "?"), ("time", "f8")])
    clf = SurvivalTree(max_depth=1, random_state=0)
    clf.fit(X, y)
    fig, ax = plt.subplots()
    anns = plot_tree_patched(clf, ax=ax)
    assert isinstance(anns, list)
    plt.close(fig)


import types


class DummyTree:
    def __init__(
        self,
        n_classes=2,
        n_outputs=1,
        value=None,
        impurity=None,
        threshold=None,
        feature=None,
        children_left=None,
        children_right=None,
        n_node_samples=None,
        weighted_n_node_samples=None,
    ):
        self.n_classes = [n_classes]
        self.n_outputs = n_outputs
        self.value = value if value is not None else np.array([[[10, 5]]])
        self.impurity = impurity if impurity is not None else np.array([0.5])
        self.threshold = threshold if threshold is not None else np.array([0.5])
        self.feature = feature if feature is not None else np.array([0])
        self.children_left = children_left if children_left is not None else np.array([-1])
        self.children_right = children_right if children_right is not None else np.array([-1])
        self.n_node_samples = n_node_samples if n_node_samples is not None else np.array([15])
        self.weighted_n_node_samples = (
            weighted_n_node_samples if weighted_n_node_samples is not None else np.array([15.0])
        )


def test_get_color_and_fill_color_classification():
    exporter = _MPLTreeExporter()
    exporter.colors = {"rgb": _color_brew(2), "bounds": None}
    tree = DummyTree()
    color = exporter.get_color([0.7, 0.3])
    assert isinstance(color, str) and color.startswith("#")
    fill = exporter.get_fill_color(tree, 0)
    assert isinstance(fill, str) and fill.startswith("#")


def test_get_color_and_fill_color_regression():
    exporter = _MPLTreeExporter()
    exporter.colors = {"rgb": _color_brew(1), "bounds": (0, 1)}
    tree = DummyTree(n_classes=1, value=np.array([[[0.5]]]))
    color = exporter.get_color(0.5)
    assert isinstance(color, str) and color.startswith("#")
    fill = exporter.get_fill_color(tree, 0)
    assert isinstance(fill, str) and fill.startswith("#")


def test_node_to_str_various_options():
    exporter = _MPLTreeExporter(node_ids=True, label="all", proportion=True)
    exporter.colors = {"rgb": _color_brew(2), "bounds": None}
    exporter.characters = ["#", "[", "]", "<=", "\n", "", ""]
    tree = DummyTree(
        n_classes=2,
        value=np.array([[[10, 5]]]),
        impurity=np.array([0.5]),
        feature=np.array([0]),
        threshold=np.array([0.5]),
        children_left=np.array([-1]),
        n_node_samples=np.array([15]),
        weighted_n_node_samples=np.array([15.0]),
    )
    s = exporter.node_to_str(tree, 0, criterion="gini")
    assert isinstance(s, str)

    # Test with regression
    tree_reg = DummyTree(n_classes=1, value=np.array([[[0.5]]]), impurity=np.array([0.2]))
    s2 = exporter.node_to_str(tree_reg, 0, criterion="mse")
    assert isinstance(s2, str)


def test_node_to_str_invalid_node_plot():
    exporter = _MPLTreeExporter()
    exporter.colors = {"rgb": _color_brew(1), "bounds": None}
    exporter.node_plot = "invalid"
    tree = DummyTree(n_classes=1, value=np.zeros((1, 1, 1)), impurity=np.array([0.1]))
    # Simulate logrank criterion and label
    with pytest.raises(ValueError):
        exporter.node_to_str(tree, 0, criterion="logrank")


def test_node_to_str_label_root():
    exporter = _MPLTreeExporter(node_ids=True, label="root")
    exporter.colors = {"rgb": _color_brew(2), "bounds": None}
    exporter.characters = ["#", "[", "]", "<=", "\n", "", ""]
    tree = DummyTree(
        n_classes=2,
        value=np.array([[[10, 5]]]),
        impurity=np.array([0.5]),
        feature=np.array([0]),
        threshold=np.array([0.5]),
        children_left=np.array([-1]),
        n_node_samples=np.array([15]),
        weighted_n_node_samples=np.array([15.0]),
    )
    s = exporter.node_to_str(tree, 0, criterion="gini")
    assert isinstance(s, str)



import warnings
import numpy as np
from scipy import stats
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap

from .visualization_extra import _input_validation, compute_max_visual_len
from .visualization_extra import define_relative_position, plot_arrow
from .utilities import frmt_pretty_print


def plot_rules(
    rules,
    preds,
    baselines,
    weights,
    max_rulelen=None,
    other_preds=None,
    preds_distr=None,
    b_box_pred=None,
    conf_level=None,
    tot_digits=4,
    cmap="RdYlGn_r",
    base_fontsize=13,
):
    """
    A visualization tool for BELLATREX, a local random forest explainability
    toolbox.

    @param rules: A list of lists, where each inner list contains strings
        representing the decision rules that are taken.
    @param preds: A list of lists, of the same shape as `rules`, where each
        inner list contains numbers representing the prediction at each point
        of the rule path.
    @param baselines: A list indicating the baseline prediction for each rule.
    @param weights: A list indicating the weight of each rule.
    @param max_rulelen: Maximum number of rules shown for each decision path.
    @param other_preds: Optional list of lists containing `preds` for other
        trees in the random forest.
    @param preds_distr: Optional list of predictions made by the random forest
        on a set of training/testing patients. Determines the x-limits for the density plot
    @param conf_level: Optional float, it greys out the interval corresponding to the confidence level
    @param tot_digits: Default: 4. number of significative digits to show
    @param cmap: The colormap used for visualization. Use "RdYlGn_r" if lower
        predictions is better. Omit the "_r" if the reverse holds.
    @param b_box_pred: Optional float (or list of) with prediction of the
        original black-box model, for the sake of comparison
    @return: List of figure and axes objects, for further customizing the graph.
    """

    # Validate inputs and determine maximum rule length
    _input_validation(rules, preds, baselines, weights)

    max_rulelen_visual = compute_max_visual_len(rules, max_rulelen=max_rulelen)

    nrules = len(rules)

    for i in range(nrules):
        assert len(rules[i]) == len(preds[i])
        if len(rules[i]) > max_rulelen_visual:
            # +1 because we need to replace the last one
            omitted = len(rules[i]) - max_rulelen_visual + 1
            rules[i][max_rulelen_visual - 1] = f"+{omitted} other rule splits"
            preds[i][max_rulelen_visual - 1] = preds[i][-1]
            rules[i] = rules[i][:max_rulelen_visual]
            preds[i] = preds[i][:max_rulelen_visual]

    if other_preds is not None:  # sets up distribution of all rule esitmates
        for i, other_pred in enumerate(other_preds):
            if len(other_pred) > max_rulelen_visual:
                other_pred[max_rulelen_visual - 1] = other_pred[-1]  # check leaf node
                other_pred = other_pred[:max_rulelen_visual]

    if preds_distr is not None:
        density = stats.gaussian_kde(preds_distr)

        extent = preds_distr.max() - preds_distr.min()
        x = np.linspace(preds_distr.min() - 0.00 * extent, preds_distr.max() + 0.00 * extent, 100)

    # Make a colorpicker

    if cmap is None:
        cmap = "RdYlGn_r"

    if cmap != "shap":
        cmap = plt.get_cmap(cmap)
    else:
        shap_blue = "#008bfb"  # SHAP blue
        shap_red = "#ff0051"
        # Create the colormap
        cmap = LinearSegmentedColormap.from_list("shap_cmap", [shap_blue, shap_red], N=256)

    deviations = [np.array(preds[i]) - baselines[i] for i in range(nrules)]

    dev_min = min(np.min(dev) for dev in [0, *deviations])  # include 0 if all positives
    dev_max = max(np.max(dev) for dev in [0, *deviations])  # include 0 if all negatives

    # choose best colorbar to taste:
    # norm = plt.matplotlib.colors.Normalize(vmin=-dev_max, vmax=dev_max)
    norm = plt.matplotlib.colors.Normalize(
        vmin=-0.5 * (dev_max - dev_min), vmax=0.5 * (dev_max - dev_min)
    )

    get_color = lambda value, baseline: cmap(norm(value - baseline))

    # Initialize the plot (rules and arrows only)
    plot_height_rulebased = 0.9 * max(max_rulelen_visual, 4)

    if preds_distr is None:  # no extra axis objects for density plot
        fig, aaxs = plt.subplots(
            figsize=(5 * nrules + 2, plot_height_rulebased + 2),
            nrows=2,
            ncols=nrules,
            sharey=True,
            gridspec_kw={"hspace": 0, "height_ratios": [plot_height_rulebased, 1]},
        )
        # rule_axs = np.atleast_1d(aaxs)
        if len(aaxs.shape) == 1:
            aaxs = np.atleast_2d(aaxs).T
        rule_axs = aaxs[0, :]
        dens_axs = [None] * len(rule_axs)  # useful placeholder for later iterations (?)

    else:  # create extra axis object for density plots (2d array)
        fig, aaxs = plt.subplots(
            figsize=(5 * nrules + 2, plot_height_rulebased + 3),
            nrows=3,
            ncols=nrules,
            sharey="row",
            gridspec_kw={"hspace": 0, "height_ratios": [plot_height_rulebased, 1, 1]},
        )
        if len(aaxs.shape) == 1:
            aaxs = np.atleast_2d(aaxs).T
        rule_axs = aaxs[0, :]
        dens_axs = aaxs[1, :]

    # Prepare to set minimum and maximum deviation as xaxis limits
    margin = 0.01 * (dev_max - dev_min)  # 1% margin left and right
    min_rel_x_axis = np.min(baselines) + dev_min - margin
    max_rel_x_axis = np.max(baselines) + dev_max + margin

    if preds_distr is not None:  # extend axis to fit underlying preds_distr min and max values
        max_rel_x_axis = max(max_rel_x_axis, x[-1])
        min_rel_x_axis = min(min_rel_x_axis, x[0])

    # set the y_label only on the leftmost plot (saves horizontal space)
    rule_axs[0].set_ylabel("Rule depth", fontsize=base_fontsize)

    for i, ax in enumerate(rule_axs):

        ax.invert_yaxis()
        # make the x axis include all partial prediction of the forest internal nodes
        # plus some margin, given by the few lines above. We choose to use
        # the same x-axis limits for all plots, for a better interpretability
        ax.set_xlim([min_rel_x_axis, max_rel_x_axis])
        ax.set_ylim([max_rulelen_visual + 0.75, -0.75])
        ax.set_yticks(range(max_rulelen_visual + 1))
        ax.tick_params(axis="y", labelsize=base_fontsize)
        ax.grid(axis="x", zorder=-99, alpha=0.5)
        ax.set_title(
            f"Selected rule {i+1}\n (weight = {100*weights[i]:.0f}%)", fontsize=base_fontsize
        )  # (weighted {weights[i]:.2f})")

    plt.subplots_adjust(wspace=0.12)
    # alt: max_rulelen_visual --> fig.get_size_inches()[0]
    aspect = 20 * (max_rulelen_visual / 5)  # because aspect=20 is ideal when max_rulelen_visual=5
    cbar = plt.colorbar(
        plt.cm.ScalarMappable(norm=norm, cmap=cmap),
        ax=aaxs,
        pad=0.04,
        aspect=aspect,
        label="Change w.r.t. baseline",
    )
    cbar.ax.tick_params(labelsize=base_fontsize + 1)
    cbar.set_label(
        "Change w.r.t. baseline", fontsize=base_fontsize + 1
    )  # Adjusting the main label size (again)

    # Visualize the entire forest
    if other_preds:
        for bsl, ax in zip(baselines, rule_axs):
            for pred in other_preds:
                ax.plot(
                    [*bsl, *pred],
                    np.arange(len(pred) + 1),
                    c=[0.9, 0.9, 0.9],
                    alpha=1.0,
                    zorder=-500,
                    lw=0.6,
                )  # zorder compared to zorder in rule_axs

    # Highlight the rule of interest on each plot
    for bsl, rule, pred, ax in zip(baselines, rules, preds, rule_axs):
        traj = [*bsl, *pred]
        bsl = bsl[0]  # structure has changed upstream and bsl is now a single-element list
        pad = 0.3
        xmin, xmax = ax.get_xlim()

        xtext_base_block = max(bsl, 0.85 * xmin + 0.15 * xmax)
        xtext_base_block = min(xtext_base_block, 0.15 * xmin + 0.85 * xmax)

        ax.text(
            s=f"Baseline\n{frmt_pretty_print(traj[0], tot_digits)}",
            fontsize=base_fontsize,
            x=xtext_base_block,
            y=-pad,
            ha="center",
            va="center",
            bbox=dict(boxstyle=f"square,pad={pad}", fc="w", ec="k", alpha=0.5),
        )

        isRight = pred[-1] < bsl
        ha = ["left", "right"][isRight]
        xtext_pred_block = max(pred[-1], 0.7 * xmin + 0.3 * xmax)
        xtext_pred_block = min(xtext_pred_block, 0.3 * xmin + 0.7 * xmax)

        ax.text(
            s=f"Prediction\n{frmt_pretty_print(pred[-1], tot_digits)}",
            fontsize=base_fontsize,
            x=xtext_pred_block,
            y=len(pred) + pad,
            ha=ha,
            va="center",
            bbox=dict(boxstyle=f"square,pad={pad}", fc="w", ec="k", alpha=0.5, zorder=5),
        )

        for j, _ in enumerate(rule):
            color = get_color(pred[j], bsl)
            # Draw the arrow
            ax.annotate(
                text="",
                xy=(traj[j + 1], j + 1),
                xytext=(traj[j], j),
                arrowprops=dict(
                    arrowstyle="-|>",
                    linewidth=2,
                    shrinkB=0,
                    mutation_scale=20,
                    edgecolor=color,
                    facecolor=color,
                ),
            )
            # Draw text, aligned betweeb the arrow trajectory and the center of the plot
            # xmin, xmax = ax.get_xlim()
            xtext = (
                2 * traj[j] + 2 * traj[j + 1] + 6 * (xmax + xmin) / 2
            ) / 10  # traj[0] originally

            closest = np.argmin([xtext - xmin, np.abs(xtext - (xmin + xmax) / 2), xmax - xtext])
            ha = ["left", "center", "right"][closest]
            ax.text(
                s=parse(rule[j]),
                x=xtext,
                y=j + 1 / 2.1,
                ha=ha,
                va="center",
                fontsize=base_fontsize,
                bbox=dict(boxstyle="square,pad=0", fc="w", ec="w", lw=1, alpha=0.75),
            )

    # Draw the distribution (density) on each rule-plot
    if preds_distr is not None:
        # Training set distribution (as provided by preds_distr)
        for i, (bsl, pred, ax) in enumerate(zip(baselines, preds, dens_axs)):

            # avg_bsl = np.mean(bsl, axis=0)
            avg_bsl = np.mean(baselines[i])  # assumes single-class output
            assert isinstance(avg_bsl, (float, int))

            # Set the density plot axes limits to match the rule plot axes limits
            ax.set_xlim(rule_axs[i].get_xlim())
            # Plot density plot with dot marker on top of the curve and dottes lines around
            ax.plot(x, density(x), "k")
            col1 = "gray"  # get_color(bsl     , bsl)
            col2 = get_color(pred[-1], avg_bsl)
            ax.plot(avg_bsl, density(avg_bsl), ".", c=col1, ms=15)
            ax.plot(pred[-1], density(pred[-1]), ".", c=col2, ms=15)
            ax.vlines(x=avg_bsl, ymin=0, ymax=density(avg_bsl), colors=col1, linestyles=":")
            ax.vlines(x=pred[-1], ymin=0, ymax=density(pred[-1]), colors=col2, linestyles=":")
            ax.set_ylim([0, 1.1 * ax.get_ylim()[1]])
            ax.set_yticks([])
            ax.grid(axis="x", zorder=-99, alpha=0.5)
            ax.set_xlabel("Prediction", fontsize=base_fontsize - 1)
            ax.tick_params(axis="x", labelsize=base_fontsize)

        dens_axs[0].set_ylabel("Density", fontsize=base_fontsize)

        # Connect density plot axes to the rest of the plot
        for bsl, pred, ax, dax in zip(baselines, preds, rule_axs, dens_axs):

            avg_bsl = np.mean(bsl, axis=0)
            # Draw dotted vline from baseline to density
            ax.vlines(x=bsl, ymin=0, ymax=ax.get_ylim()[0], colors="gray", linestyles=":")
            dax.vlines(
                x=bsl, ymin=density(avg_bsl), ymax=dax.get_ylim()[1], colors="gray", linestyles=":"
            )
            # Draw dotted line from final prediction to density
            ax.vlines(
                x=pred[-1],
                ymin=max_rulelen_visual,
                ymax=ax.get_ylim()[0],
                colors=get_color(pred[-1], avg_bsl),
                linestyles=":",
            )
            dax.vlines(
                x=pred[-1],
                ymin=density(pred[-1]),
                ymax=dax.get_ylim()[1],
                colors=get_color(pred[-1], avg_bsl),
                linestyles=":",
            )

    ### Here confidence interval:

    if other_preds is None and conf_level is not None:
        warnings.warn("No confidsence level can be plotted if 'other_preds' is None")

    # Do so by plotting distribution integrated in the preds_distr
    if other_preds is not None and conf_level is not None and dens_axs[0] is not None:

        final_preds = [pred[-1] for pred in other_preds] + [pred[-1] for pred in preds]

        # conf_level:  size of the confidence interval (0.9 = 90% of the tree predictions)
        if conf_level <= 0 or conf_level > 1:
            raise ValueError("conf_level must be in the (0, 1] range. Found:", conf_level)

        vmin = np.quantile(final_preds, (1 - conf_level) / 2)
        vmax = np.quantile(final_preds, (1 + conf_level) / 2)

        for ax in aaxs[-2, :]:  # last aaxs row is for the arrows, we need the axis just above them
            ax.fill_between(
                x=[vmin, vmax],
                y1=[0, 0],
                y2=[ax.get_ylim()[1], ax.get_ylim()[1]],
                color="gray",
                alpha=0.2,
            )

    ## Add weighted average contribution to the bottom of the plot on the arrow Axes object
    n_cols = aaxs.shape[1]
    # define relative position of aroow subplot depending on n_cols
    pos_list = define_relative_position(n_cols)

    rule_preds = [preds[i][-1] for i in range(len(rules))]
    xlim_values = [ax.get_xlim() for ax in aaxs[0, :]]

    # Assing xlim values of arrowplot (aaxs[-1,:]) to be
    # equal to the xlim values of the ruleplot (aaxs[0,:])
    for ax, xlim in zip(aaxs[-1, :], xlim_values):
        ax.set_xlim(xlim)

    for j, pos in zip(range(n_cols), pos_list):

        # TODO: make class out of all this so that plot_arrow can inherit
        # attribtues from plot_rules and/or instances of BellatrexExplain()
        aaxs[-1, j] = plot_arrow(
            aaxs[-1, j],
            pos,
            weight=weights[j],
            pred_out=rule_preds[j],
            fontsize=base_fontsize,
            tot_digits=tot_digits,
        )

    # Add final prediction to the plot
    final_pred = np.sum([weights[i] * preds[i][-1] for i in range(len(rules))])
    final_pred_str = frmt_pretty_print(final_pred, tot_digits)
    final_pred_str = f"Bellatrex weighted prediction: {final_pred_str}"
    final_pred_str += (
        " (= "
        + " + ".join(
            [
                rf"{frmt_pretty_print(preds[i][-1], tot_digits)}" rf"$\times${weights[i]:.2f}"
                for i in range(len(rules))
            ]
        )
        + ")"
        + " " * 12
    )
    # extra spaces to the right to move text to the left

    if b_box_pred is not None:
        ypos_fin_text = 0.05
        bbox_pred_str = frmt_pretty_print(
            b_box_pred, tot_digits
        )  # works both with single values and multi-output
        bbox_pred_str = f"\n(compared to black-box model prediction: {bbox_pred_str})      "
        # bbox_pred_str += ", ".join([f"{frmt_pretty_print(pred)}"
        #                              for pred in np.atleast_1d(b_box_pred)])
        # bbox_pred_str += ")"
    else:
        ypos_fin_text = 0.08
        bbox_pred_str = ""

    # plt.figtext(0.5, ypos_fin_text, final_pred_str+bbox_pred_str, fontsize=base_fontsize+2, ha="center")
    fig.text(
        0.5, ypos_fin_text, final_pred_str + bbox_pred_str, fontsize=base_fontsize + 2, ha="center"
    )

    return fig, aaxs


def parse(rulesplit):
    """Parses a rulesplit outputted by Bellatrex into a form suitable for visualization."""

    # 1) Replace special characters by LaTeX symbols
    rulesplit = rulesplit.replace("≤", r"$\leq$")
    rulesplit = rulesplit.replace("<=", r"$\leq$")
    rulesplit = rulesplit.replace("≥", r"$\geq$")
    rulesplit = rulesplit.replace(">=", r"$\geq$")

    # 2) Remove information related to the current value, situated after the threshold value (right hand side)
    end_math_idx = rulesplit.rfind("$") + 1
    if end_math_idx and end_math_idx < len(rulesplit):
        # Get the substring after the last closing dollar sign
        right_substring = rulesplit[end_math_idx:]
        # Check and process for "(" in the substring after the last $
        if "(" in right_substring:
            # Find the last "(" in the substring and adjust it relative to the whole string
            last_paren_index = right_substring.rfind("(") + end_math_idx

            # Update rulesplit without the part after the last "("
            rulesplit = rulesplit[:last_paren_index].strip()

    return rulesplit


def read_rules(file, file_extra=None):
    """
    Reads and parses rules, predictions, baselines, and weights from a given file,
    with an optional additional file for extra predictions.

    The function processes the content of the specified file to extract rule weights, baseline predictions, rules,
    and corresponding predictions. If an additional file is provided, it extracts extra predictions as well.

    Args:
        file (str): The path to the primary file containing rules, weights, baselines, and predictions.
        file_extra (str, optional): The path to the additional file containing extra predictions. Defaults to None.

    Returns:
        tuple: A tuple containing:
            - f_rules (list): A list of rules, where each rule is a list of conditions.
            - f_preds (list): A list of predictions corresponding to each rule.
            - f_baselines (list): A list of baseline predictions.
            - f_weights (list): A list of weights for each rule.
            - other_preds (list or None): List of unused rules stored in the -extra file,
                if provided; otherwise, None.
    """
    f_rules = []
    f_preds = []
    f_baselines = []
    f_weights = []
    with open(file, "r", encoding="utf-8") as f:
        btrex_rules = f.readlines()
    for line in btrex_rules:
        if "RULE WEIGHT" in line:
            f_weights.append(float(line.split(":")[1].strip("\n").strip(" #")))
        if "Baseline prediction" in line:
            # Handle multiple comma-separated baseline predictions
            baseline_values = line.split(":")[1].strip(" \n").split(",")
            f_baselines.append(
                [float(value.strip()) for value in baseline_values]
            )  # Append as a list
            rule = []
            pred = []
        if "node" in line:
            fullrule = line.split(":")[1].strip().strip("\n").split("-->")
            index_thresh = max([fullrule[0].find(char) for char in ["=", "<", ">"]])
            fullrule[0] = fullrule[0][0 : index_thresh + 8]
            rule.append(fullrule[0])
            # Handle multiple comma-separated predictions
            predictions = fullrule[1].split(",")
            pred.extend([float(p.strip()) for p in predictions])
        if "leaf" in line:
            f_rules.append(rule)
            f_preds.append(pred)

    if file_extra:
        other_preds = []
        with open(file_extra, "r", encoding="utf-8") as f:
            btrex_extra = f.readlines()
        for line in btrex_extra:
            if "Baseline prediction" in line:
                pred = []
            if "node" in line:
                pred_values = line.split("-->")[1].strip().split(",")
                pred.extend([float(p.strip()) for p in pred_values])
            if "leaf" in line:
                other_preds.append(pred)
    else:
        other_preds = None

    # f_rules: list of lists, where each inner list contains decision splits strings as elements
    # f_preds: list of lists, same shape and structure as f_rules, where every element
    #       is a float with node-level predictions (or list for multi-output trees, in the future)
    # f_baselines: list of lists, every element contaains baseline (root) predictions.
    #   Single-element list for single output trees.
    # f_weights: list of floats, every element is a weight of the corresponding (tree) rule
    # other_preds: list of lists, same shape and structure as f_preds, but for unused rules
    #   Results in a much longer list of rules (the unused rules) comapred to f_rules.

    return f_rules, f_preds, f_baselines, f_weights, other_preds



import numpy as np

# from tree_extraction import TreeExtraction
# from TreeRepresentation import Jaccard_trees
# from TreeRepresentation import Jaccard_rules
from .tree_representation_utils import tree_splits_to_vector, rule_splits_to_vector


class TreeDissimilarity:

    def __init__(
        self,
        clf,
        set_up,
        tree_indeces,
        dissim_method="rules",
        feature_represent="weighted",
        sample=None,
    ):
        self.clf = clf
        self.set_up = set_up
        self.tree_indeces = tree_indeces
        self.dissim_method = dissim_method
        self.feature_represent = feature_represent
        self.sample = sample

    def return_avg_dissim_and_matrix(self):
        """
        This is the main function that computes everything we need. TODO move out of the class?
        """

        vectors = self.tree_to_vectors(
            self.clf, self.tree_indeces, self.dissim_method, self.feature_represent, self.sample
        )

        diss_matrix = self.vectors_to_dissim_matrix(vectors)
        avg_dissimilarity = self.compute_avg_dissimilarity(diss_matrix)

        return avg_dissimilarity, diss_matrix

    def tree_to_vectors(self, clf, tree_indeces, dissim_method, feature_represent, sample):
        vectors = []  # list of the vector representation ( tree or path)

        if tree_indeces is None:  # in case no indeces are given, use all the trees of the ensemble
            tree_indeces = range(clf.n_estimators)

        if dissim_method == "trees":
            for idx in tree_indeces:
                # weights "by sample size" or "simple" cases are considered
                vectors.append(tree_splits_to_vector(clf, idx, feature_represent))

        if dissim_method == "rules":
            for idx in tree_indeces:
                # weights "by sample size" or "simple" cases are considered
                vectors.append(rule_splits_to_vector(clf, idx, feature_represent, sample))
        return vectors

    def vectors_to_dissim_matrix(self, vector_list: list):

        size = len(vector_list)
        A = np.zeros([size, size])

        for i in range(size):
            for j in range(i, size):
                A[i, j] = np.sum(np.minimum(vector_list[i], vector_list[j])) / np.sum(
                    np.maximum(vector_list[i], vector_list[j])
                )
                A[j, i] = A[i, j]

        return 1 - A  # dissimilarity instead of similarity (A[i,i]= 0)

    def compute_avg_dissimilarity(self, dist_matrix: np.ndarray):
        # averages the OFF_DIAGONAL elements of the matrix
        # if matrix is (1 x 1) return np.nan (division by zero!)
        if dist_matrix.shape[0] != dist_matrix.shape[1]:
            raise ValueError(f"Expected a square matrix, but got shape: {dist_matrix.shape}")
        if dist_matrix.shape[0] > 1 and dist_matrix.shape[1] > 1:
            return dist_matrix.sum() / dist_matrix.shape[0] / (dist_matrix.shape[1] - 1)
        # In a 2x2 matrix, the average is computed over a single element taken twice, then / 2
        else:
            return np.nan



"""
Author: Klest Dedja
Here we manually test most of the the pipeline, from data loading to model explanation building.
We cover most of the combinations of tasks, models, and settings.
"""

import pytest
import matplotlib

matplotlib.use("Agg")  # Must be before importing pyplot

import matplotlib.pyplot as plt  # Safe after backend is set
from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sksurv.ensemble import RandomSurvivalForest

from bellatrex import BellatrexExplain
from bellatrex.utilities import get_auto_setup

# from bellatrex.wrapper_class import pack_trained_ensemble
from bellatrex.datasets import (
    load_mlc_data,
    load_regression_data,
    load_survival_data,
    load_binary_data,
    load_mtr_data,
)

import os

IS_CI = os.environ.get("CI") == "true"

MAX_TEST_SAMPLES = 2

DATA_LOADERS = {
    "binary": load_binary_data,
    "regression": load_regression_data,
    "survival": load_survival_data,
    "multi-label": load_mlc_data,
    "multi-target": load_mtr_data,
}


# --- Main setup logic shared by the other tests ---
def prepare_fitted_bellatrex(setup, loader):
    X, y = loader(return_X_y=True)
    X_train, X_test, y_train, _ = train_test_split(X, y, test_size=0.3, random_state=0)
    assert setup == get_auto_setup(
        y
    ), f"Automatic task detection failed: found {get_auto_setup(y)} instead of {setup}"

    if setup.lower() == "survival":
        clf = RandomSurvivalForest(
            n_estimators=100, min_samples_split=10, n_jobs=-2, random_state=0
        )
    elif setup.lower() in ["binary", "multi-label"]:
        clf = RandomForestClassifier(
            n_estimators=100, min_samples_split=5, n_jobs=-2, random_state=0
        )
    elif setup.lower() in ["regression", "multi-target"]:
        clf = RandomForestRegressor(
            n_estimators=100, min_samples_split=5, n_jobs=-2, random_state=0
        )
    else:
        raise ValueError(f"Detection task {setup} not compatible with Bellatrex (yet)")

    clf.fit(X_train, y_train)
    test_grid = {"n_trees": [0.6, 1.0], "n_dims": [2, None], "n_clusters": [1, 2, 3]}
    btrex_fitted = BellatrexExplain(clf, set_up="auto", p_grid=test_grid, verbose=3).fit(
        X_train, y_train
    )

    return btrex_fitted, X_test


# --- Core (non-GUI) test ---
def test_core_workflow():

    for setup, loader in DATA_LOADERS.items():  # Iterate over loading functions
        btrex_fitted, X_test = prepare_fitted_bellatrex(setup, loader)
        for i in range(MAX_TEST_SAMPLES):
            tuned_method = btrex_fitted.explain(X_test, i)
            tuned_method.plot_overview(show=not IS_CI, plot_gui=False)


# --- Rules and file handling test ---
def test_create_rules_txt():
    for setup, loader in DATA_LOADERS.items():
        X, y = loader(return_X_y=True)
        X_train, X_test, y_train, _ = train_test_split(X, y, test_size=0.3, random_state=0)

        if setup.lower() == "survival":
            clf = RandomSurvivalForest(
                n_estimators=100, min_samples_split=5, n_jobs=-1, random_state=0
            )
        elif setup.lower() in ["binary", "multi-label"]:
            clf = RandomForestClassifier(
                n_estimators=100, min_samples_split=5, n_jobs=-1, random_state=0
            )
        elif setup.lower() in ["regression", "multi-target"]:
            clf = RandomForestRegressor(
                n_estimators=100, min_samples_split=5, n_jobs=-1, random_state=0
            )
        else:
            raise ValueError(f"Detection task {setup} not compatible with Bellatrex (yet)")

        clf.fit(X_train, y_train)
        btrex_fitted = BellatrexExplain(clf, set_up="auto", verbose=1).fit(X_train, y_train)

        for i in range(MAX_TEST_SAMPLES):
            btrex_fitted.explain(X_test, i)
            out_file = "test_rules.txt"
            out_file, out_file_extra = btrex_fitted.create_rules_txt(
                out_dir="explanations-output", out_file=out_file
            )
            assert os.path.exists(out_file), "Rules file was not created"

            with open(out_file, "r", encoding="utf8") as f:
                content = f.read()
                assert len(content) > 0, "Rules file is empty."

            enough_trees = clf.n_estimators * min(btrex_fitted.p_grid["n_trees"]) > max(
                btrex_fitted.p_grid["n_clusters"]
            )
            os.remove(out_file)
            # If there are more trees than clusters, check for extra rules file and remove it
            if enough_trees:

                assert os.path.exists(out_file_extra), "Rules file was not created"
                with open(out_file_extra, "r", encoding="utf8") as f:
                    content = f.read()
                    assert len(content) > 0, "Rules file is empty."

                os.remove(out_file_extra)


# --- GUI test with plot_gui=True ---
# @pytest.mark.gui
# def test_gui_workflow_old():
#     if IS_CI:
#         matplotlib.use("Agg")  # Non-blocking backend when running in CI

#     for _, loader in DATA_LOADERS.items():  # Iterate over loading functions
#         btrex_fitted, X_test = prepare_fitted_bellatrex(loader)

#         for i in range(MAX_TEST_SAMPLES):
#             tuned_method = btrex_fitted.explain(X_test, i)
#             fig, obj = tuned_method.plot_overview(show=not IS_CI, plot_gui=True)
#             if not IS_CI: # if show=True
#                 assert fig is not None
#                 plt.close(fig)
#             else:
#                 assert obj is not None  # DearPyGui plot objects


@pytest.mark.gui
def test_gui_workflow():

    dpg = pytest.importorskip(
        "dearpygui.dearpygui", reason="Install Bellatrex[gui] to run GUI tests"
    )

    for setup, loader in DATA_LOADERS.items():
        btrex_fitted, X_test = prepare_fitted_bellatrex(setup, loader)

        for i in range(MAX_TEST_SAMPLES):
            tuned_method = btrex_fitted.explain(X_test, i)

            # Enforce plot_gui=False in headless mode
            fig, obj = tuned_method.plot_overview(show=not IS_CI, plot_gui=True)
            assert obj is not None
            plt.close(fig)



import os
import warnings
import math

import dearpygui.dearpygui as dpg
from dearpygui_ext.themes import create_theme_imgui_light

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter
import matplotlib.cm as cm
from sklearn.decomposition import PCA

from .utilities import colormap_from_str
from .plot_tree_patch import plot_tree_patched
from .utilities import rule_print_inline, custom_axes_limit
from .utilities import custom_formatter

IS_CI = os.environ.get("CI") == "true"


class InteractPoint:  # Object containg all information of a point
    def __init__(self, name, pos, color, size, shape, cluster_memb=None, value=None):
        self.name = str(name)  # string
        self.pos = pos  # list or tuple with two floats
        self.color = color  # list or tuple with 4 integers from 0 to 255 (rgba)
        # self.edges=edges                   #string (default: "black"), customisation not implemented
        self.size = size  # positive integer
        self.shape = shape  # string
        self.cluster_memb = cluster_memb  # string or None
        self.value = value  # float or None (learner predicted value)


class InteractPlot:  # Object containing all information of a plot
    def __init__(self, name, points, clustered=False, xlabel="PC1", ylabel="PC2"):
        self.name = str(name)  # string (should be unique)
        self.points = points  # list of InteractPoints objects
        self.clustered = clustered  # Boolean
        self.xlabel = xlabel  # string
        self.ylabel = ylabel  # string


def make_interactive_plot(plots, temp_files_dir, plot_size=700, other_inputs=None, max_depth=None):
    """
    Generates an interactive Graphical User Interface with (default) two plots, offering the
    the possibility to interact with them.

    Parameters:
    -----------
    plots : list
        A list of interactable_plot objects that contain the points and plotting information.
    plot_size : int, optional, default=700
        The size of the plot in pixels.
    other_inputs : object, optional
        An object containing additional inputs for the user, such as a fitted classifier and sample data.
    max_depth : int, optional
        The maximum depth for displaying the decision tree.

    Returns:
    --------
    list
        A list of interactable_plot objects for further handling and testing.

    """
    border_size = 0.1  # Proportion of borders around the plot.
    dpg.destroy_context()
    dpg.create_context()

    # DPG set theme
    lighttheme = create_theme_imgui_light()
    dpg.bind_theme(lighttheme)

    border_size = 0.1  # Proportion of borders around the plot.
    fullspacing = 8
    # dpg.destroy_context()
    # dpg.create_context()

    # Load and set up tree image
    # create image to load:
    # Activated when mouse is left clicked on a DPG object
    # only at this stage, the tree index is known,
    # as it corresponds to the point.hovered at the moment
    def mouse_click_left_callback():

        # Check all points to see if they are hovered
        for plot in plots:
            for point in plot.points:
                if point.hovered:
                    assert other_inputs is not None
                    # the other_inputs is supposed to be a TreeExtractor instance

                    my_clf = (
                        other_inputs.clf
                    )  # rf estimator here. Is this compatible with the Wrapper? Probably not
                    if hasattr(my_clf, "feature_names_in_"):
                        feature_names = my_clf.feature_names_in_
                    else:
                        feature_names = [f"X{i}" for i in range(my_clf.n_features_in_)]
                    sample_index = other_inputs.sample.index[0]
                    tree_index = int(point.name)
                    the_tree = my_clf[tree_index]

                    rule_print_inline(the_tree, other_inputs.sample)

                    tree_name_png = "Tree_" + str(tree_index) + ".png"

                    if max_depth is not None:
                        real_plot_leaves = max(the_tree.tree_.n_leaves, 2 ** (max_depth - 1))
                        real_plot_depth = min(the_tree.tree_.max_depth, max_depth)
                    else:
                        real_plot_leaves = the_tree.tree_.n_leaves
                        real_plot_depth = the_tree.tree_.max_depth

                    smart_width = 1 + 0.4 * real_plot_leaves
                    smart_width = int(smart_width)
                    smart_height = int(real_plot_depth + 1)
                    if (
                        my_clf.n_outputs_ > 3
                    ):  # create extra vertical space for multi-output trees, since output is printed over multiple lines
                        smart_height = int(smart_height * (0.92 + 0.08 * (my_clf.n_outputs_)))

                    plt.subplots(figsize=(smart_width, smart_height))

                    plot_tree_patched(
                        the_tree, max_depth=max_depth, feature_names=feature_names, fontsize=8
                    )

                    plt.title(
                        "Tree %i for sample index %i" % (tree_index, sample_index),
                        fontsize=10 + int(1.3 * real_plot_depth),
                    )
                    plt.savefig(os.path.join(temp_files_dir, tree_name_png))

                    if plt.isinteractive():
                        plt.show()  # Show selected tree in console, if interactive
                    # now that the needed tree image is created, we can load it
                    # generates provisional plot: load tree_name_png, or "trial" file:
                    # "PL"+plot.name+"_"+point.name+".png"
                    IMwidth, IMheight, _, IMdata = dpg.load_image(
                        os.path.join(temp_files_dir, tree_name_png)
                    )
                    os.remove(os.path.join(temp_files_dir, tree_name_png))

                    with dpg.texture_registry(show=False):
                        treeplot = dpg.add_static_texture(IMwidth, IMheight, IMdata)

                    popup_window = dpg.add_window(
                        width=min(IMwidth + 4 * fullspacing, windowwidth),
                        height=min(IMheight + 5 * fullspacing, windowheight),
                        modal=True,
                        horizontal_scrollbar=True,
                    )

                    dpg.add_image(treeplot, width=IMwidth, height=IMheight, parent=popup_window)

    # DPG mouse click setup
    with dpg.handler_registry():
        dpg.add_mouse_click_handler(
            button=dpg.mvMouseButton_Left, callback=mouse_click_left_callback
        )

    # Activated every frame ## IS THIS EVEN USED? CHECK!
    def point_draw_and_interact(sender, app_data, user_data):

        # Calculates positions in pixel space
        _helper_data = app_data[0]
        transformed_x = app_data[1]
        transformed_y = app_data[2]
        mouse_x_pixel_space = _helper_data["MouseX_PixelSpace"]
        mouse_y_pixel_space = _helper_data["MouseY_PixelSpace"]

        # Renaming for legibility
        plot = user_data
        points = plot.points

        # Deletes the previous frame
        # Slot = 2 specifies that only the dpg.draw objects must be deleted.
        dpg.delete_item(sender, children_only=True, slot=2)
        dpg.configure_item(sender, tooltip=False)

        # Goes through all points to draw them and to see if they are hovered
        for i, point in enumerate(points):

            # Draws point shape
            if point.shape == "circle":
                dpg.draw_circle(
                    [transformed_x[i], transformed_y[i]],
                    point.size,
                    fill=point.color,
                    parent=sender,
                )

            elif point.shape == "triangle":

                sin_size = 0.500 * point.size
                cos_size = 0.866 * point.size

                dpg.draw_triangle(
                    [transformed_x[i], transformed_y[i] - point.size],
                    [transformed_x[i] + cos_size, transformed_y[i] + sin_size],
                    [transformed_x[i] - cos_size, transformed_y[i] + sin_size],
                    fill=point.color,
                    parent=sender,
                )

            elif point.shape == "square":
                half_diagonal = 0.707 * point.size
                dpg.draw_rectangle(
                    [transformed_x[i] - half_diagonal, transformed_y[i] - half_diagonal],
                    [transformed_x[i] + half_diagonal, transformed_y[i] + half_diagonal],
                    fill=point.color,
                    parent=sender,
                )

            elif point.shape == "star":

                N_VERTECES = 5
                outer_radius = point.size * 1.1
                inner_radius = outer_radius * 0.4

                # Calculate the radial cooridnates for each vertex of the star
                outer_angles = [
                    math.radians(360 / N_VERTECES * i - 90 / N_VERTECES) for i in range(N_VERTECES)
                ]
                inner_angles = [
                    math.radians(360 / N_VERTECES * i + 90 / N_VERTECES) for i in range(N_VERTECES)
                ]

                # Calculate the coordinates of each vertex
                vertices = []
                for j in range(N_VERTECES):
                    outer_vertex = [
                        transformed_x[i] + outer_radius * math.cos(outer_angles[j]),
                        transformed_y[i] + outer_radius * math.sin(outer_angles[j]),
                    ]
                    inner_vertex = [
                        transformed_x[i] + inner_radius * math.cos(inner_angles[j]),
                        transformed_y[i] + inner_radius * math.sin(inner_angles[j]),
                    ]
                    vertices.extend([outer_vertex, inner_vertex])

                # Draw the star using the calculated vertices
                dpg.draw_polygon(vertices, fill=point.color, parent=sender)

            else:
                raise ValueError(f"Shape {point.shape} not supported")

            # Checks if cursor close enough for point to be considered hovered
            x_dif = transformed_x[i] - mouse_x_pixel_space
            y_dif = transformed_y[i] - mouse_y_pixel_space
            norm_squared = x_dif**2 + y_dif**2
            if norm_squared <= point.size**2:
                point.hovered = True
                dpg.configure_item(sender, tooltip=True)

                # Adds extra information to tooltip depending on point type
                tool_tip_text = "hovered Point: " + point.name
                if point.cluster_memb is not None:
                    tool_tip_text += "\n" + "Cluster: " + point.cluster_memb
                if point.value is not None:
                    tool_tip_text += "\n" + "Value: " + point.value

                # Shows tooltip when point is hovered
                dpg.set_value("TooltipText" + plot.name, tool_tip_text)
            else:
                point.hovered = False

    # Function that draws the average position of the different clusters
    def cluster_average(sender, app_data):
        size = 9
        x = app_data[1]
        y = app_data[2]

        # Reset previous frame
        dpg.delete_item(
            sender, children_only=True, slot=2
        )  # Deletes all drawn items (draw = slot 2)

        # Draws the average position
        for i in range(len(x)):
            dpg.draw_circle([x[i], y[i]], size, fill=[150, 150, 150, 255], parent=sender)
            dpg.draw_text([x[i] - size / 2, y[i] - size], clusters[i], size=size * 2, parent=sender)

    # A few useful extra variables
    plotamount = len(plots)
    addedwidth = 0
    # Now load images and calculate sizes and limits
    # Load in the temp_colourbar images, saving the desired size.

    for plot in plots:
        IMwidth, IMheight, _, IMdata = dpg.load_image(
            os.path.join(temp_files_dir, "temp_colourbar" + plot.name + ".png")
        )
        with dpg.texture_registry(show=False):
            dpg.add_static_texture(IMwidth, IMheight, IMdata, tag="temp_colourbar" + plot.name)
        plot.colour_bar_width = round(plot_size * IMwidth / IMheight)
        addedwidth += plot.colour_bar_width

    windowwidth = plotamount * plot_size + (2 * plotamount + 1) * fullspacing + addedwidth
    windowheight = plot_size + 2 * fullspacing

    # Calculates x- and y-axis limits based on points
    for plot in plots:
        x_values = [point.pos[0] for point in plot.points]
        y_values = [point.pos[1] for point in plot.points]
        x_diff = max(x_values) - min(x_values)
        y_diff = max(y_values) - min(y_values)
        plot.x_axis_limits = [
            min(x_values) - x_diff * border_size,
            max(x_values) + x_diff * border_size,
        ]
        plot.y_axis_limits = [
            min(y_values) - y_diff * border_size,
            max(y_values) + y_diff * border_size,
        ]

    sample_index = other_inputs.sample.index[0]

    # DPG build windows
    with dpg.window(
        tag="MainWindow",
        width=windowwidth,
        label="Explaining sample "
        + str(sample_index)
        + ", close the window to continue with code execution",
        height=windowheight,
        no_title_bar=False,
        no_move=False,
        no_resize=True,
    ):
        with dpg.group(horizontal=True):
            for plot in plots:
                with dpg.plot(
                    tag="Tree representation plot" + plot.name, height=plot_size, width=plot_size
                ):
                    # create x-axis
                    dpg.add_plot_axis(dpg.mvXAxis, label=plot.xlabel, tag="x_axis" + plot.name)
                    dpg.set_axis_limits(
                        "x_axis" + plot.name, plot.x_axis_limits[0], plot.x_axis_limits[1]
                    )

                    # create y-axis
                    dpg.add_plot_axis(dpg.mvYAxis, label=plot.ylabel, tag="y_axis" + plot.name)
                    dpg.set_axis_limits(
                        "y_axis" + plot.name, plot.y_axis_limits[0], plot.y_axis_limits[1]
                    )

    viewport = dpg.create_viewport(title="Plot", width=1000, height=720)
    dpg.configure_viewport(
        viewport, height=windowheight + 5 * fullspacing, width=windowwidth + 2 * fullspacing
    )

    return plots


def plot_with_interface(
    plot_data_bunch,
    kmeans,
    input_method,  # a fitted bellatrex instance
    temp_files_dir,
    max_depth=None,
    colormap=None,
    clusterplots=(True, False),
):

    def shaper(in_shape):
        if in_shape is True:
            return "star"
        else:
            return "circle"

    def sizer(in_size):
        if in_size is True:
            return 16.0
        else:
            return 9.0

    # def edger(in_edge):
    #     return "black"
    # else:
    #     return (1,1,1,0.5) # semi-transparent white

    def rgbaconv(mpl_rgba):
        dpg_rgba = [i * 255 for i in mpl_rgba]
        return dpg_rgba

    # repeat PCA to 2 dimensions for projected trees
    # (original proj dimension can be > 2)
    # better: use original one and keep only the first two dims.
    PCA_fitted = PCA(n_components=2).fit(plot_data_bunch.proj_data)
    plottable_data = PCA_fitted.transform(plot_data_bunch.proj_data)  # (lambda,2)

    # centers = PCA_fitted.transform(kmeans.cluster_centers_)
    cluster_memb = kmeans.labels_

    final_ts_idx = input_method.final_trees_idx

    is_final_candidate = [
        plot_data_bunch.index[i] in final_ts_idx for i in range(len(plot_data_bunch.index))
    ]
    custom_sizes = list(map(sizer, is_final_candidate))
    custom_shapes = list(map(shaper, is_final_candidate))
    # custom_edges = list(map(edger, is_final_candidate))

    plots = []

    color_map_left = cm.viridis  # default colormap for clustering plot
    color_map_left = mpl.colors.LinearSegmentedColormap.from_list(
        "Custom cmap", [color_map_left(i) for i in range(color_map_left.N)], color_map_left.N
    )

    color_map_right = colormap_from_str(colormap)  # User custom colormap from object or from string

    # colormap = color_map_left if clustered == True else color_map_right

    for plotindex, clustered in enumerate(clusterplots):

        if clustered:
            fig, ax = plt.subplots(1, 1, figsize=(1, 4.5), dpi=120)

            # define the bins and normalize
            freqs = np.bincount(cluster_memb)
            if np.min(freqs) == 0:
                raise KeyError(
                    "There are empty clusters, the scatter and colorbar could differ in color shade"
                )
            norm_bins = list(np.cumsum(freqs))
            norm_bins.insert(0, 0)

            # transform list to array (ticks location needs arithmentic computation)
            norm_bins = np.array(norm_bins)

            # create label names
            labels = []
            for i in np.unique(cluster_memb):
                labels.append("cl.{:d}".format(i + 1))
            # normalizing color, prepare ticks, labels
            norm = mpl.colors.BoundaryNorm(norm_bins, color_map_left.N)
            tickz = norm_bins[:-1] + (norm_bins[1:] - norm_bins[:-1]) / 2

            cb1 = mpl.colorbar.Colorbar(
                ax,
                cmap=color_map_left,
                norm=norm,
                spacing="proportional",
                ticks=tickz,
                boundaries=norm_bins,
                format="%1i",
            )
            # label="cluster membership")
            cb1.ax.set_yticklabels(labels)  # vertically oriented colorbar

            ax.yaxis.set_ticks_position("left")
            norms = [norm(norm_bins[cluster_memb[i]]) for i in range(len(cluster_memb))]

        else:  # not the clustering plot, but the prediction/loss distribution:
            fig, ax = plt.subplots(1, 1, figsize=(1.2, 4.5), dpi=110)

            # binary, regression, survival case (single-output cases)
            if isinstance(plot_data_bunch.rf_pred, float) or plot_data_bunch.rf_pred.size == 1:

                is_binary = plot_data_bunch.set_up == "bin"

                plot_data_bunch.rf_pred = np.array(plot_data_bunch.rf_pred).squeeze()

                v_min, v_max = custom_axes_limit(
                    np.array(plot_data_bunch.pred).min(),
                    np.array(plot_data_bunch.pred).max(),
                    plot_data_bunch.rf_pred,
                    is_binary,
                )

                norm_preds = mpl.colors.BoundaryNorm(
                    np.linspace(v_min, v_max, 256), color_map_right.N
                )

                ## add to colorbar a line corresponding to Bellatrex prediction
                # TODO: improve customFormatter function (ability to choose nb digits) and use that instead of np.round
                pred_tick = np.round(plot_data_bunch.rf_pred, 3)

                cb2 = mpl.colorbar.Colorbar(
                    ax, cmap=color_map_right, norm=norm_preds, label="RF pred: " + str(pred_tick)
                )

                plot_data_bunch.pred = np.array(
                    plot_data_bunch.pred
                ).squeeze()  # force shape: (n_trees,)

                # add tick in correspondence to the single trees prediction
                cb2.ax.plot([0, 1], [plot_data_bunch.pred] * 2, color="grey", linewidth=1)

                # add indicator (visually: >--<) in correspondence to the ensemble prediction
                cb2.ax.plot([0.02, 0.98], [pred_tick] * 2, color="black", linewidth=2.5, marker="P")

            # multi-output case, L2 losses instead of predictions
            else:
                # Blue for small losses, red for big losses
                # color_map_right = colormap_from_str('RdYlBu_r')
                color_map_right = colormap_from_str(
                    colormap
                )  # User custom colormap from object or from string

                v_min, v_max = custom_axes_limit(
                    np.array(plot_data_bunch.loss).min(),
                    np.array(plot_data_bunch.loss).max(),
                    force_in=None,
                    is_binary=False,
                )

                norm_preds = mpl.colors.BoundaryNorm(
                    np.linspace(v_min, v_max, 256), color_map_right.N
                )

                cb2 = mpl.colorbar.Colorbar(
                    ax,
                    cmap=color_map_right,
                    norm=norm_preds,
                    label=str(input_method.fidelity_measure) + " loss",
                )
                cb2.ax.plot([0, 1], [plot_data_bunch.loss] * 2, color="grey", linewidth=1)

            # ax.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.2g'))
            ticks_to_plot = ax.get_yticks()

            if np.abs(np.min(ticks_to_plot)) < 1e-3 and np.abs(np.max(ticks_to_plot)) > 1e-2:
                min_index = np.argmin(ticks_to_plot)
                ticks_to_plot[min_index] = 0
                ax.set_yticks(ticks_to_plot)

            ax.yaxis.set_major_formatter(FuncFormatter(custom_formatter))
            ax.minorticks_off()

            # single output case: float, or (1,)-shaped ndarray
            if isinstance(plot_data_bunch.rf_pred, float) or plot_data_bunch.rf_pred.size == 1:
                norms = [
                    norm_preds(float(plot_data_bunch.pred[i])) for i in range(len(cluster_memb))
                ]
            else:  # multi output case
                norms = [norm_preds(plot_data_bunch.loss[i]) for i in range(len(cluster_memb))]

        fig.tight_layout()
        fig.savefig(os.path.join(temp_files_dir, f"temp_colourbar{plotindex}.png"))
        plt.close(fig)  # prevent them from being shown in the console.

        # use appropriate colormap (right plot vs left plot)
        cmap_gui = color_map_left if clustered else color_map_right

        colours = [rgbaconv(cmap_gui(norms[i])) for i in range(len(plot_data_bunch.index))]

        points = []
        for j, index in enumerate(plot_data_bunch.index):
            points.append(
                InteractPoint(
                    index,
                    plottable_data[j],  # name and position
                    colours[j],
                    custom_sizes[j],  # size
                    # custom_edges[j],  # edge
                    custom_shapes[j],
                )  # shape
            )

            # adds extra information to the point for the tooltip
            if clustered:
                points[j].cluster_memb = str(
                    cluster_memb[j] + 1
                )  # +1 to change from 0-indexed to 1-indexed
            else:
                if isinstance(plot_data_bunch.pred[j], float):
                    points[j].value = (
                        f"{plot_data_bunch.pred[j]:.3f}"  # Show 3 numbers after decimal point
                    )
                elif isinstance(plot_data_bunch.loss[j], float):
                    points[j].value = (
                        f"{plot_data_bunch.loss[j]:.3f}"  # Show 3 numbers after decimal point
                    )
                else:
                    raise ValueError("expecting float, got {type(plot_data_bunch.loss)} instead")

        plots.append(InteractPlot(plotindex, points, clustered=clustered))

    # end of the enumerate(clusterplots) `for' loop (length 2 -> two plots)
    viewport_dict = make_interactive_plot(
        plots, temp_files_dir, plot_size=700, other_inputs=input_method, max_depth=max_depth
    )

    try:
        os.remove(os.path.join(temp_files_dir, "temp_colourbar0.png"))
        os.remove(os.path.join(temp_files_dir, "temp_colourbar1.png"))
    except (FileNotFoundError, UserWarning) as e:
        warnings.warn(
            f"{e}\nCould not delete temporary files correctly. Double check the "
            f"folder:\n {temp_files_dir}"
        )

    return plots  # Return the plots object for further handling



import pytest
import numpy as np
import pandas as pd
from bellatrex import tree_representation_utils as tr_utils


# Minimal DummyCSR for .toarray() compatibility
class DummyCSR:
    def toarray(self):
        return np.ones(3)

    def __array__(self, dtype=None):
        return np.ones(3)


# DummyTree for single learner and ensemble
class DummyTree:
    def __init__(self, n_features=3):
        self.feature = np.array([0, 1, -1])
        self.n_node_samples = np.array([10, 5, 5])
        self.decision_path_called = False
        self.n_features_in_ = n_features
        self.tree_ = self  # For compatibility with code expecting .tree_

    def decision_path(self, X):
        self.decision_path_called = True
        return DummyCSR()


# DummyClf for ensemble case
class DummyClf:
    def __init__(self, n_features=3):
        self.estimators_ = [DummyTree(n_features)]
        self.n_features_in_ = n_features
        self.tree_ = self.estimators_[0]  # For single-tree compatibility

    def __getitem__(self, idx):
        return self.estimators_[idx]


# 1) add_emergency_noise
@pytest.mark.filterwarnings("ignore:MDS matrix has rank 0")
def test_add_emergency_noise():
    mat = np.zeros((3, 3))
    noisy = tr_utils.add_emergency_noise(mat, noise_level=1e-2)
    assert noisy.shape == mat.shape
    assert not np.allclose(noisy, mat)


# 2) tree_splits_to_vector
def test_tree_splits_to_vector_simple():
    clf = DummyClf()
    vec = tr_utils.tree_splits_to_vector(clf, 0, split_weight="simple")
    assert isinstance(vec, np.ndarray)
    assert vec.shape[0] == clf.n_features_in_


def test_tree_splits_to_vector_by_samples():
    clf = DummyClf()
    vec = tr_utils.tree_splits_to_vector(clf, 0, split_weight="by_samples")
    assert isinstance(vec, np.ndarray)
    assert vec.shape[0] == clf.n_features_in_


def test_tree_splits_to_vector_invalid():
    clf = DummyClf()
    with pytest.raises(KeyError):
        tr_utils.tree_splits_to_vector(clf, 0, split_weight="invalid")


def test_tree_splits_to_vector_no_splits():
    # All features are -1, so the_splits is empty
    class NoSplitTree(DummyTree):
        def __init__(self, n_features=3):
            super().__init__(n_features)
            self.feature = np.array([-1, -1, -1])

    class NoSplitClf(DummyClf):
        def __init__(self, n_features=3):
            self.estimators_ = [NoSplitTree(n_features)]
            self.n_features_in_ = n_features
            self.tree_ = self.estimators_[0]

        def __getitem__(self, idx):
            return self.estimators_[idx]

    clf = NoSplitClf()
    vec = tr_utils.tree_splits_to_vector(clf, 0, split_weight="simple")
    assert isinstance(vec, np.ndarray)
    assert np.all(vec == 0)
    assert vec.shape[0] == clf.n_features_in_


def test_tree_splits_to_vector_padding():
    # n_features_in_ > number of splits, triggers padding logic
    class PadTree(DummyTree):
        def __init__(self, n_features=6):
            super().__init__(n_features)
            self.feature = np.array([0, 1, -1, 2, -1, -1])

    class PadClf(DummyClf):
        def __init__(self, n_features=6):
            self.estimators_ = [PadTree(n_features)]
            self.n_features_in_ = n_features
            self.tree_ = self.estimators_[0]

        def __getitem__(self, idx):
            return self.estimators_[idx]

    clf = PadClf()
    vec = tr_utils.tree_splits_to_vector(clf, 0, split_weight="simple")
    assert isinstance(vec, np.ndarray)
    assert vec.shape[0] == clf.n_features_in_
    # Only first few elements should be nonzero, rest should be zero
    assert np.all(vec[3:] == 0)


# 3) rule_splits_to_vector
def test_rule_splits_to_vector_simple():
    clf = DummyClf()
    sample = np.array([1, 2, 3])
    vec = tr_utils.rule_splits_to_vector(clf, 0, feature_represent="simple", sample=sample)
    assert isinstance(vec, np.ndarray)
    assert vec.shape[0] == clf.n_features_in_


def test_rule_splits_to_vector_weighted():
    clf = DummyClf()
    sample = np.array([1, 2, 3])
    vec = tr_utils.rule_splits_to_vector(clf, 0, feature_represent="weighted", sample=sample)
    assert isinstance(vec, np.ndarray)
    assert vec.shape[0] == clf.n_features_in_


def test_rule_splits_to_vector_invalid():
    clf = DummyClf()
    sample = np.array([1, 2, 3])
    with pytest.raises(KeyError):
        tr_utils.rule_splits_to_vector(clf, 0, feature_represent="invalid", sample=sample)



import pytest
from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sksurv.ensemble import RandomSurvivalForest

from bellatrex.bellatrex_explain import BellatrexExplain
from bellatrex.datasets import (
    load_mlc_data,
    load_regression_data,
    load_survival_data,
    load_binary_data,
    load_mtr_data,
)

DATA_LOADERS = {
    "binary": load_binary_data,
    "regression": load_regression_data,
    "survival": load_survival_data,
    "multi-label": load_mlc_data,
    "multi-target": load_mtr_data,
}

MAX_TEST_SAMPLES = 2

# --- Edge Case Test: Single Data Point ---
# def test_single_data_point():
#     for loader in DATA_LOADERS:
#         X, y = loader(return_X_y=True)
#         X_single, y_single = X[:1], y[:1]  # Single data point
#         SETUP = get_auto_setup(y_single)

#         if SETUP.lower() in "survival":
#             clf = RandomSurvivalForest(n_estimators=10, min_samples_split=2, n_jobs=-1, random_state=0)
#         elif SETUP.lower() in ["binary", "multi-label"]:
#             clf = RandomForestClassifier(n_estimators=10, min_samples_split=2, n_jobs=-1, random_state=0)
#         elif SETUP.lower() in ["regression", "multi-target"]:
#             clf = RandomForestRegressor(n_estimators=10, min_samples_split=2, n_jobs=-1, random_state=0)
#         else:
#             raise ValueError(f"Detection task {SETUP} not compatible with Bellatrex (yet)")

#         clf.fit(X_single, y_single)
#         btrex_fitted = BellatrexExplain(clf, set_up="auto", verbose=3).fit(X_single, y_single)
#         explanation = btrex_fitted.explain(X_single, 0)
#         assert explanation is not None, "Explanation failed for single data point"


# --- Error Handling Tests ---
def test_invalid_hyperparameters():
    for setup, loader in DATA_LOADERS.items():
        X, y = loader(return_X_y=True)
        X_train, _, y_train, _ = train_test_split(X, y, test_size=0.3, random_state=0)

        if setup.lower() in "survival":
            clf = RandomSurvivalForest(
                n_estimators=10, min_samples_split=5, n_jobs=-1, random_state=0
            )
        elif setup.lower() in ["binary", "multi-label"]:
            clf = RandomForestClassifier(
                n_estimators=10, min_samples_split=5, n_jobs=-1, random_state=0
            )
        elif setup.lower() in ["regression", "multi-target"]:
            clf = RandomForestRegressor(
                n_estimators=10, min_samples_split=5, n_jobs=-1, random_state=0
            )
        else:
            raise ValueError(f"Detection task {setup} not compatible with Bellatrex (yet)")

        clf.fit(X_train, y_train)

        invalid_p_grid = {"n_trees": [-1, 0], "n_dims": [0, 1], "n_clusters": [0, "invalid"]}
        with pytest.raises(ValueError):
            BellatrexExplain(clf, set_up="auto", p_grid=invalid_p_grid).fit(X_train, y_train)


def test_unfitted_model():
    clf = RandomForestClassifier(n_estimators=10, random_state=0)
    btrex = BellatrexExplain(clf, set_up="auto")
    with pytest.raises(Exception):
        btrex.explain(None, 0)


def test_unsupported_model():
    class UnsupportedModel:
        """Just a dummy class to simulate an unsupported model type."""
        pass

    clf = UnsupportedModel()
    try:
        BellatrexExplain(clf, set_up="auto")
    except ValueError as e:
        assert "Unsupported model type" in str(e), "Expected ValueError for unsupported model"



# -*- coding: utf-8 -*-
"""
Created on Mon Oct  9 14:55:55 2023

@author:       Klest Dedja
@institution:  KU Leuven
"""
import warnings
import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix, hstack

from sklearn.utils.validation import check_is_fitted

from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sksurv.tree import SurvivalTree
from sksurv.ensemble import RandomSurvivalForest


def pack_trained_ensemble(clf, set_up="auto", time_to_bin=None):
    """
    Packs a trained ensemble model into a dictionary format compatible with scikit-learn.

    Parameters
    ----------
    clf : RandomForestRegressor, RandomForestClassifier, RandomSurvivalForest
        The trained ensemble classifier to be packed. The classifier must be one of the supported types:
        RandomForestRegressor, RandomForestClassifier, or RandomSurvivalForest.

    set_up : str, optional, default="auto"
        Specifies the format in which the tree data should be output. This argument is passed to the
        `tree_to_dict` function. The default value is "auto".

    time_to_bin : float or int, optional
        This parameter is specific to `RandomSurvivalForest`. If provided, it will be passed to the
        `tree_to_dict` function for each tree. If `time_to_bin` is provided for any other classifier
        type, a warning will be issued, and the parameter will be ignored.

    Returns
    -------
    dict
        A dictionary representation of the packed ensemble model. This format is compatible with scikit-learn's
        model loading functions.

    Raises
    ------
    ValueError
        If the `clf` object is not one of the supported classifier types.

    Notes
    -----
    This function first checks if the classifier is fitted using the `check_is_fitted` method.
    The individual trees of the ensemble are then converted into a dictionary format using the
    `tree_to_dict` function and stored in a list. Finally, the list of tree dictionaries is
    converted into a model dictionary using `tree_list_to_model`.
    """

    if not isinstance(clf, (RandomForestRegressor, RandomForestClassifier, RandomSurvivalForest)):
        raise ValueError(f"Incompatible classifier object, found: {type(clf)}")

    check_is_fitted(clf)  # if not fitted, exception is raised

    if time_to_bin is not None and not isinstance(clf, RandomSurvivalForest):
        warnings.warn(
            f"time_to_bin parameter is ignored for any estimator other than RandomSurvivalForest.\n"
            f"Found: {type(clf)}"
        )

    tree_list = []
    # Correct access to n_estimators
    if hasattr(clf, 'n_estimators'):
        for t in range(clf.n_estimators):
            tree_dict = tree_to_dict(clf, t, output_format=set_up, time_to_bin=time_to_bin)
            tree_list.append(tree_dict)
    else:
        raise ValueError(f"Unsupported classifier type '{type(clf)}'.")

    # load the model in a dict format somewhat compatible with scikit-learn:
    clf_out = tree_list_to_model(tree_list)

    return clf_out


class EnsembleWrapper:
    """
    This class serves as a wrapper for compatibility with
    tree ensemble models stored as a list of dictionaries.
    see "Example of loading a custom tree model into SHAP" in
    https://shap.readthedocs.io/en/stable/tabular_examples.html
    (link verified as of July 2024)
    It is designed to be compatible with:
        RandomForestClassifier,
        RandomForestRegressor,
        RandomSurvivalForest.
    It assumes that the predictions are AVERAGED in the ensemble step
    """

    class Estimator:

        class Tree_:
            def __init__(
                self,
                feature,
                n_leaves,
                n_node_samples,
                children_left,
                children_right,
                threshold,
                value,
                max_depth,
                feature_names_in_,
                n_features_in_,
                learner_class,
            ):
                self.feature = feature
                self.n_leaves = n_leaves
                self.n_node_samples = n_node_samples
                self.children_left = children_left
                self.children_right = children_right
                self.threshold = threshold
                self.value = value
                self.max_depth = max_depth
                self.feature_names_in_ = feature_names_in_
                self.n_features_in_ = n_features_in_
                self.learner_class = learner_class

            @property
            def n_outputs_(self):
                return len(self.value[0])

            @property
            def node_count(self):
                return len(self.feature)

            def apply(self, X):
                n_samples = X.shape[0]
                node_indices = []
                for i in range(n_samples):
                    node_indices.append(self._apply_tree(0, X[i, :]))
                return node_indices

            def _apply_tree(self, node_index, sample):
                if self.children_left[node_index] == -1 and self.children_right[node_index] == -1:
                    return [node_index]

                path_indices = [node_index]
                if sample[self.feature[node_index]] <= self.threshold[node_index]:
                    path_indices.extend(self._apply_tree(self.children_left[node_index], sample))
                else:
                    path_indices.extend(self._apply_tree(self.children_right[node_index], sample))

                return path_indices

            def decision_path(self, X):
                if isinstance(X, pd.DataFrame):
                    X = X.to_numpy()
                node_indices_list = self.apply(X)
                data, indices, indptr = [], [], [0]
                for sample_nodes in node_indices_list:
                    data.extend([1] * len(sample_nodes))
                    indices.extend(sample_nodes)
                    indptr.append(len(indices))
                return csr_matrix((data, indices, indptr), shape=(X.shape[0], self.node_count))

        # this is the __init__ of the Estimator class (intermediate layer)
        def __init__(self, tree_dict):
            self.tree_ = self.Tree_(  # ORDER IS IMPORTANT! Check how class Tree_ is initialized
                # The .get method returns None if key is missing
                tree_dict.get("features"),
                tree_dict.get("n_leaves"),
                tree_dict.get("node_sample_weight"),
                tree_dict.get("children_left"),
                tree_dict.get("children_right"),
                tree_dict.get("thresholds"),
                tree_dict.get("values"),  # Does it need to be spelled as singular? .value?
                tree_dict.get("max_depth"),
                tree_dict.get("feature_names_in_"),
                tree_dict.get("n_features_in_"),
                tree_dict.get("learner_class"),
            )
            self.n_outputs_ = self.tree_.n_outputs_  # Inferred from underlying tree_.
            self.learner_class = self.tree_.learner_class

        def predict(self, X):
            if isinstance(X, pd.DataFrame):
                X = X.to_numpy()
            node_indices = self.tree_.apply(X)
            n_samples = X.shape[0]
            predictions = np.zeros((n_samples, self.n_outputs_))

            for i in range(n_samples):
                leaf_idx = node_indices[i][-1]  # Assuming the last node is the leaf
                predictions[i, :] = self.tree_.value[leaf_idx]

            return predictions

        def decision_path(self, X):
            return self.tree_.decision_path(X)

    # init EnsembleWrapper here (most external class)
    def __init__(self, clf_dict):
        self.estimators_ = [self.Estimator(tree) for tree in clf_dict["trees"]]
        self.n_outputs_ = self.estimators_[0].n_outputs_
        self.n_estimators = len(self.estimators_)
        self.feature_names_in_ = clf_dict["trees"][0]["feature_names_in_"]
        self.n_features_in_ = clf_dict["trees"][0]["n_features_in_"]
        self.ensemble_class = clf_dict["ensemble_class"]

        if "unique_times_" in clf_dict["trees"][0]:
            self.unique_times_ = clf_dict["trees"][0]["unique_times_"]

    def __getitem__(self, index):
        return self.estimators_[index]

    def predict(self, X):  # the output format depends on what was stored in the original dict
        if isinstance(X, pd.DataFrame):
            X = X.to_numpy()

        n_samples = X.shape[0]
        ensemble_predictions = np.zeros((n_samples, self.n_outputs_))

        for estimator in self.estimators_:  # sum over learners, stored dictionaries assume this
            ensemble_predictions += estimator.predict(X)

        return ensemble_predictions / self.n_estimators  # assuming NON additive ensemble here

    def decision_path(self, X):

        if isinstance(X, pd.DataFrame):
            X = X.to_numpy()

        all_paths = []
        n_nodes_ptr = [0]

        for estimator in self.estimators_:
            path_csr = estimator.decision_path(X)
            all_paths.append(path_csr)
            # Handle None case for path_csr
            if path_csr is not None:
                n_nodes_ptr.append(n_nodes_ptr[-1] + path_csr.shape[1])
            else:
                n_nodes_ptr.append(n_nodes_ptr[-1])  # Handle None case gracefully
                warnings.warn(
                    "Decision path is None for one of the estimators,  this may have unintendend consequences."
                )
        all_paths_csr = hstack(all_paths).tocsr()

        return all_paths_csr, np.array(n_nodes_ptr)


def tree_to_dict(clf_obj, idx, output_format, time_to_bin=None):
    """
    Compatible with single output trees only, at the moment.
    compatible with SurvivalTree learners of a RandomSurvivalForest
    with DecisionTreeClassifier and DecisionTreeRegressor from RandomForests
    """
    tree_obj = clf_obj[idx]
    tree = tree_obj.tree_

    tree_dict = {
        "children_left": tree.children_left,
        "children_right": tree.children_right,
        "children_default": tree.children_right.copy(),  # to be changed when sklearn can handle missing values
        "n_leaves": tree.n_leaves,
        "max_depth": tree.max_depth,
        "features": tree.feature,
        "thresholds": tree.threshold,
        "node_sample_weight": tree.weighted_n_node_samples,
        "feature_names_in_": getattr(clf_obj, "feature_names_in_", None),
        "n_features_in_": getattr(tree_obj, "n_features_in_", None),
        "unique_times_": getattr(tree_obj, "unique_times_", None),
        "is_event_time_": getattr(tree_obj, "is_event_time_", None),
        "random_state": getattr(tree_obj, "random_state", None),
        "ensemble_class": clf_obj.__class__.__name__,
        "learner_class": tree_obj.__class__.__name__,
    }

    tree_dict["output_format"] = output_format

    if isinstance(tree_obj, SurvivalTree): # case SETUP=survival

        if tree_dict["unique_times_"] is None and output_format not in ["probability"]:
            raise KeyError("Missing 'unique_times_' in the tree ensemble.")

        if (
            time_to_bin is None and tree_dict["unique_times_"] is not None
        ):  # select median time interval
            time_to_bin = tree_dict["unique_times_"][len(tree_dict["unique_times_"]) // 2]

        if output_format == "predict_survival_curve":
            tree_dict["values"] = tree.value[:, :, 1]

        elif output_format in ["hazard", "auto"]:
            # sum CHF over unique_times_ that have an event (imitating SurvivalTree .predict)
            tree_dict["values"] = np.sum(
                tree.value[:, tree_dict["is_event_time_"], 0], axis=1
            ).reshape(-1, 1)

        elif output_format in ["survival", "time-to-event"]:

            tree_dict["values"] = np.trapz(
                tree.value[:, :, 1], tree_obj.unique_times_, axis=1
            ).reshape(
                -1, 1
            )  # integrate S(t) over unique_times_
            # output shape: (n_nodes, 1) with E[S(t)] at each node

        elif output_format == "probability":
            # pick last "False" index before "True" appears
            index_T = np.argmax(tree_obj.unique_times_ > time_to_bin) - 1
            # it DOES work when all times are > T_bin, as it will again select -1
            if min(tree_obj.unique_times_) > time_to_bin:
                index_T = 0
            # probability of experiencing the event by t=2 -> P(t) = 1 - S(t)
            tree_dict["values"] = 1 - tree.value[:, index_T, 1].reshape(-1, 1)
            # Why was the reshape needed? Why single-element arrays like this? Hmm...

        else:
            raise ValueError("Input not recognized. Double check")

    # case SETUP=binary:
    elif (
        isinstance(tree_obj, DecisionTreeClassifier)
        and tree_obj.n_outputs_ == 1
        and output_format in ["probability", "auto"]
    ):
        partials = tree.value[:, 0, :]  # output for 2 classes, now take average
        tree_dict["values"] = (partials[:, 1] / (partials[:, 0] + partials[:, 1])).reshape(-1, 1)

    elif ( #case SETUP=multi-label
        isinstance(tree_obj, DecisionTreeClassifier)
        and tree_obj.n_outputs_ > 1
        and output_format in ["probability", "auto"]
    ):
        partials = tree.value
        tree_dict["values"] = partials[:, :, 1] / (partials[:, :, 0] + partials[:, :, 1])

    elif ( # case SETUP=regression
        isinstance(tree_obj, DecisionTreeRegressor)
        and tree_obj.n_outputs_ == 1
        and output_format in ["probability", "auto"]
    ):
        tree_dict["values"] = tree.value[:, 0, 0].reshape(-1, 1)  # output (n_nodes, 1)

    elif ( # case SETUP=multi-target
        isinstance(tree_obj, DecisionTreeRegressor)
        and tree_obj.n_outputs_ > 1
        and output_format in ["probability", "auto"]
    ):
        tree_dict["values"] = tree.value[:, :, 0]  # output (n_nodes, n_outputs_)

    else:
        raise ValueError(
            f"Combination of learner '{tree_obj.__class__.__name__}' and"
            f" scenario '{output_format}' is not recognized"
        )

    tree_dict["base_offset"] = tree_dict["values"][0]  # root node (0) prediction

    return tree_dict


def tree_list_to_model(tree_list):
    """
    given each learner is stored as a dict, create a list of such dictionaries.
    NOTE that this is NOT directly compatible with the SHAP library,
    which assumes that the input custom tree models are additive

    To use SHAP, the user needs to divide the predicted values accordingly.
    See commmented lines below
    """
    # for t in tree_list: # divide here
    #     t["values"] = t["values"] / len(tree_list)

    base_offset = np.mean(np.array([t["base_offset"] for t in tree_list]))

    assert len(set([t["output_format"] for t in tree_list])) == 1  # consistency check
    assert len(set([t["ensemble_class"] for t in tree_list])) == 1  # consistency check

    output_format = tree_list[0]["output_format"]
    ensemble_class = tree_list[0]["ensemble_class"]

    model_as_dict = {
        "trees": tree_list,  # list of dicts
        "base_offset": base_offset,  # single value (average of array)
        "output_format": output_format,
        "ensemble_class": ensemble_class,
        "input_dtype": np.float32,
        "internal_dtype": np.float32,
    }

    return model_as_dict



import os
import warnings
from joblib import Parallel, delayed
import numpy as np
import pandas as pd

import matplotlib
import matplotlib.pyplot as plt

from sklearn.exceptions import ConvergenceWarning
from sklearn.model_selection import ParameterGrid
from sklearn.utils.validation import check_is_fitted
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

from sksurv.ensemble import RandomSurvivalForest

from .wrapper_class import EnsembleWrapper
from .utilities import predict_helper
from .visualization_extra import _input_validation
from .tree_extraction import TreeExtraction
from .utilities import plot_preselected_trees, rule_print_inline
from .utilities import rule_to_file, frmt_pretty_print
from .visualization import read_rules, plot_rules


class BellatrexExplain:

    FONT_SIZE = 14
    MAX_FEATURE_PRINT = 10

    def __init__(
        self,
        clf,
        set_up="auto",
        force_refit=False,
        verbose=0,
        proj_method="PCA",
        dissim_method="rules",
        feature_represent="weighted",
        p_grid={
            "n_trees": [0.6, 0.8, 1.0],
            "n_dims": [2, None],
            "n_clusters": [1, 2, 3],
        },
        pre_select_trees="L2",
        fidelity_measure="L2",
        n_jobs=1,
        ys_oracle=None,
    ):

        self.clf = clf  # (un)fitted instance of R(S)F
        self.set_up = set_up
        self.force_refit = force_refit
        self.proj_method = proj_method
        self.dissim_method = dissim_method
        self.feature_represent = feature_represent
        self.p_grid = p_grid
        self.pre_select_trees = pre_select_trees
        self.fidelity_measure = fidelity_measure
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.ys_oracle = None

    def _validate_p_grid(self):
        """
        This method validates and sets the parameters for the hyperparameter grid.
        It checks if the provided keys in the p_grid dictionary are valid,
        and sets default values if any of them are missing.
        It also checks if the values provided for n_trees are valid and raises
        errors if necessary. Finally, it converts n_trees to the number
        of trees used by the underlying ensemble model in case the values
        for n_trees are given as proportions.

        Raises:
            ValueError: If n_trees is less than or equal to 0, or if the list of n_trees contains both
                        proportions and integers, or if any n_trees value is greater than n_estimators.
            Warning: If the hyperparameter list contains unexpected keys other
            than the default set, this function reverts to using default values.
        """

        if not isinstance(self.p_grid, dict):
            raise ValueError(
                "p_grid parameter is expected to be a dictionary. Found {type(self.p_grid)} instead."
            )

        default_keys = ["n_trees", "n_dims", "n_clusters"]

        unexpected_keys = [key for key in self.p_grid.keys() if key not in default_keys]
        if unexpected_keys:
            warnings.warn(
                f"The hyperparameter list contains unexpected keys: {unexpected_keys}. Ignoring them."
            )

        if "n_trees" not in self.p_grid.keys():
            self.n_trees = [0.6, 0.8, 1.0]  # set to default if not existing
        else:
            self.n_trees = self.p_grid["n_trees"]  # CAN BE A LIST

        if "n_dims" not in self.p_grid.keys() or self.p_grid["n_dims"] is None:
            self.n_dims = [None]  # set to default if not existing
        else:
            self.n_dims = self.p_grid["n_dims"]  # CAN BE A LIST
            # treat 'all' as None (compatible with sklearn's PCA)
            self.n_dims = [None if x == "all" else x for x in self.n_dims]

        if "n_clusters" not in self.p_grid.keys():
            self.n_clusters = [1, 2, 3]  # set to default if not existing
        else:
            self.n_clusters = self.p_grid["n_clusters"]  # CAN BE A LIST

        if min(self.n_trees) <= 0:
            raise ValueError("n_trees must be all > 0")

        if min(self.n_trees) < 1.0 and max(self.n_trees) > 1.0:
            raise ValueError(
                "The list of n_trees must either indicate a proportion"
                " of trees in the (0,1] interval, or indicate the number"
                " of tree learners."
            )

        # Check that the n_trees provided by the user does not exceed the number of total trees
        # in the R(S)F. This works for both a fitted sklearn model and a dictionary

        if max(self.n_trees) > self.clf.n_estimators:
            raise ValueError("'n_trees' hyperparater value cannot be greater than n_estimators")

        # if proportion of n_trees is given instead, check correctness and transform to integer values:
        if np.array(
            [isinstance(i, float) for i in self.n_trees]
        ).all():  # if all elements are floats
            if (
                max(self.n_trees) <= 1.0 and min(self.n_trees) > 0
            ):  # all elements are in the (0, 1] interval:
                # round to closest integer
                self.n_trees = (
                    (np.array(self.n_trees) * self.clf.n_estimators + 0.5).astype(int).tolist()
                )  # all params are sorted as list, keep consistency

    def is_fitted(self):  # auxiliary function that returns boolean
        """
        This function determines whether the classifier (`self.clf`) has been fitted.
        It considers two scenarios:
        - If `self.clf` is a dictionary, it is assumed to represent a pre-trained model.
            In this case, the function wraps  the dictionary using the `EnsembleWrapper` class
            and returns `True`, indicating the model is fitted.
        - If `self.clf` is not a dictionary, the function performs a check using `check_is_fitted`
            from sklearn. This check is applicable to sklearn or sksurv models. If `check_is_fitted`
            does not raise an exception, the function returns `True`. If an exception is raised,
            the function returns `False`, indicating the model is not fitted.

        Returns:
            bool: `True` if the model is fitted, `False` otherwise.
        """

        if isinstance(self.clf, dict):  # case of simple, packed dictionary:
            self.clf = EnsembleWrapper(
                self.clf
            )  # EnsembleWrapper() ensures compatibily of (packed) dictionaries
            return True
        elif isinstance(
            self.clf, EnsembleWrapper
        ):  # case where compatibility is already taken care of:
            return True
        else:  # case where full sklearn/sksurv model is given. Check if it is fitted or not
            try:
                check_is_fitted(self.clf)  # only with sklearn models (but works with all of them)
                return True
            except Exception:  # if 'check_is_fitted' throws exception, we need it to output 'False'
                return False
        # Note that from sklearn 1.3. we can simply use return _is_fitted(self.clf) # returns boolean already

    def fit(self, X, y):
        """
        Fits the classifier to the data if not already fitted or if force refit is requested.
        It also checks the validity of the hyperparameters and sets up the prediction task
        based on the type of fitted classifier.

        Parameters:
        - X : array-like, shape (n_samples, n_features). Training dataset.
        - y : array-like, shape (n_samples,) or (n_samples, n_outputs). Target values.

        Raises:
        - ValueError: If an incompatible model type is provided or specific conditions required by
            the model setup are not met. For example, if a dictionary format is used when 'auto'
            set-up is selected, or if the classifier is not recognized/supported by the framework.

        Returns:
        - self : object
        Returns the instance itself.

        Notes:
        - If 'force_refit' is False and the model is already fitted, it skips the fitting process
            and proceeds to build an explanation.
        - If 'verbose' is set to 0 or higher, it will print the fitting status.
        - Automatically determines the prediction task ('set_up') based on classifier properties.
        """

        if self.force_refit is False and self.is_fitted():
            if self.verbose >= 0:
                print("Model is already fitted, building explanation.")
        else:
            if self.verbose >= 0:
                print("Fitting the model...", end="")
            if hasattr(self.clf, "n_jobs"):
                self.clf.n_jobs = self.n_jobs
            self.clf.fit(X, y)
            if self.verbose >= 0:
                print("fitting complete")

        # then check whether the input grid values are admissible
        self._validate_p_grid()

        if self.verbose >= 2:
            print(f"oracle_sample is: {self.ys_oracle}")

        if self.set_up == "auto":  # automatically determine scenario based on fitted classifier

            if isinstance(self.clf, EnsembleWrapper):

                if self.clf.ensemble_class == "RandomForestClassifier" and self.clf.n_outputs_ == 1:
                    self.set_up = "binary"
                elif (
                    self.clf.ensemble_class == "RandomForestClassifier" and self.clf.n_outputs_ > 1
                ):
                    self.set_up = "multi-label"
                elif (
                    self.clf.ensemble_class == "RandomForestRegressor" and self.clf.n_outputs_ == 1
                ):
                    self.set_up = "regression"
                elif self.clf.ensemble_class == "RandomForestRegressor" and self.clf.n_outputs_ > 1:
                    self.set_up = "multi-target"
                elif (
                    self.clf.ensemble_class == "RandomSurvivalForest" and y.shape[1] == 2
                ):  # ideally, check for 1 field boolean and one field numeric
                    self.set_up = "survival"
                elif self.clf.ensemble_class == "RandomSurvivalForest" and y.shape[1] > 2:
                    raise ValueError(
                        f"Shape of recarray labels {y.shape} implies multi-output survival analysis, "
                        "which is not implemented yet"
                    )
                else:
                    raise ValueError(
                        f"Classifier {self.clf.ensemble_class} not compatible "
                        "with 'auto' set-up selection. Please select the set-up manually"
                    )

            elif isinstance(self.clf, RandomForestClassifier):
                if self.clf.n_outputs_ == 1:
                    self.set_up = "binary"
                else:
                    self.set_up = "multi-label"
            elif isinstance(self.clf, RandomForestRegressor):
                if np.array(y).ndim < 2 or self.clf.n_outputs_ == 1:
                    self.set_up = "regression"
                else:
                    self.set_up = "multi-target"
            elif isinstance(self.clf, RandomSurvivalForest):
                if self.clf.n_outputs_ == self.clf.unique_times_.shape[0]:
                    self.set_up = "survival"
                else:
                    self.set_up = "multi-variate-sa"
                    raise ValueError(
                        "n_outputs_ shape != unique_times_ shape:"
                        f"{self.clf.n_outputs_.shape} != {self.clf.unique_times_.shape} \n"
                        "Note that multi-event Survival analysis is not supported yet"
                    )
            else:
                raise ValueError(
                    "Provided model is not recognized or compatible with Bellatrex:", self.clf
                )

            if self.verbose > 0:
                print(f"Automatically setting prediction task to: {self.set_up}")

        return self

    def explain(self, X, idx):

        sample = X.iloc[[idx]]

        if self.ys_oracle is not None:
            self.ys_oracle = self.ys_oracle.iloc[idx]  # pick needed one

        param_grid = {"n_trees": self.n_trees, "n_dims": self.n_dims, "n_clusters": self.n_clusters}

        for key, value in param_grid.items():
            if not isinstance(value, (list, np.ndarray)):
                param_grid[key] = [value]

        grid_list = list(ParameterGrid(param_grid))
        best_perf = -np.inf

        trees_extract = TreeExtraction(
            self.proj_method,
            self.dissim_method,
            self.feature_represent,
            self.n_trees,
            self.n_dims,
            self.n_clusters,
            self.pre_select_trees,
            self.fidelity_measure,
            self.clf,
            self.ys_oracle,
            self.set_up,
            sample,
            self.verbose,
        )

        best_params = {
            "n_clusters": 2,
            "n_dims": None,
            "n_trees": 80,
        }  # default combination, in case everything fails

        if self.n_jobs == 1:
            for params in grid_list:
                try:
                    candidate = trees_extract.set_params(**params).main_fit()
                    perf = candidate.score(self.fidelity_measure, self.ys_oracle)
                except ConvergenceWarning as e:
                    warnings.warn(f"Reached ConvergenceWarning: {e}, skipping candidate: {params}")
                    perf = -np.inf
                # except KeyboardInterrupt as e:
                #     raise e

                if self.verbose >= 5:
                    print("params:", params)
                    print(f"fidelity current candidate: {perf:.4f}")

                if perf > best_perf:
                    best_perf = perf
                    best_params = params

            if best_perf == -np.inf:
                warnings.warn(
                    "The GridSearch did not find any meaningful configuration,"
                    " setting default parameters"
                )
        elif self.n_jobs > 1:
            warnings.warn(
                "Multi-processing is not optimized in this version and the"
                " speed-up gain is marginal. Set n_jobs = 1 to avoid this warning"
            )

            def missing_params_dict(given_params, class_instance):
                param_names = class_instance.__init__.__code__.co_varnames[1:]
                param_values = {name: getattr(class_instance, name) for name in param_names}
                missing_params = {
                    key: value for key, value in param_values.items() if key not in given_params
                }
                return missing_params

            provided_params = list(grid_list[0].keys())
            class_instance = trees_extract
            constant_params = missing_params_dict(provided_params, class_instance)

            def create_btrex_candidate(constant_params, **params):
                return TreeExtraction(**constant_params, **params)

            def run_candidate(
                create_instance_func, fidelity_measure, ys_oracle, constant_params, **params
            ):
                candidate = None
                try:
                    etrees_instance = create_instance_func(constant_params, **params)
                    candidate = etrees_instance.main_fit()
                    perf = candidate.score(fidelity_measure, ys_oracle)
                except Exception as e:
                    warnings.warn(f"Something went wrong ({e}), skipping candidate: {params}")
                    perf = -np.inf
                return perf, params

            results = Parallel(n_jobs=self.n_jobs, prefer="threads")(
                delayed(run_candidate)(
                    create_btrex_candidate,
                    self.fidelity_measure,
                    self.ys_oracle,
                    constant_params,
                    **params,
                )
                for params in grid_list
            )

            perfs, params_list = zip(*results)

            if best_perf > -np.inf:
                best_idx = np.argsort(perfs)[::-1][0]
                best_perf = perfs[best_idx]
                best_params = params_list[best_idx]

            if best_perf == -np.inf:
                warnings.warn(
                    "The GridSearch did not find any functioning hyperparameter"
                    " configuration, setting default configuration"
                )

        tuned_method = trees_extract.set_params(**best_params).main_fit()
        tuned_method.sample_score = tuned_method.score(self.fidelity_measure, self.ys_oracle)

        # Ensure final_trees_idx and cluster_sizes are not None and are arrays
        if tuned_method.final_trees_idx is not None:
            final_extract_trees = np.array(tuned_method.final_trees_idx)
        else:
            final_extract_trees = np.array([])
        if tuned_method.cluster_sizes is not None:
            final_cluster_sizes = np.array(tuned_method.cluster_sizes)
        else:
            final_cluster_sizes = np.array([])

        if not isinstance(self.clf, RandomSurvivalForest):
            surrogate_pred = np.array([0.0] * self.clf.n_outputs_).reshape(sample.shape[0], -1)
        else:  # Assumed RSF is only single-output
            surrogate_pred = np.array([0.0])

        for tree_idx, cluster_size in zip(final_extract_trees, final_cluster_sizes):
            if len(final_cluster_sizes) > 0:
                cluster_weight = cluster_size / np.sum(final_cluster_sizes)
            else:
                cluster_weight = 0  # Default value

            surrogate_pred += predict_helper(self.clf[tree_idx], sample.values) * cluster_weight

        # Store surrogate prediction directly in BellatrexExplain instance
        # self.surrogate_prediction = surrogate_pred
        surrogate_pred_str = frmt_pretty_print(surrogate_pred, digits_single=4)

        if self.verbose >= 1:
            print("best params:", best_params)
            print(f"Achieved fidelity: {best_perf:.4f}")

        if self.verbose >= 2:
            print(f"final trees indices: {final_extract_trees}")
            print(f"final cluster sizes: {final_cluster_sizes}")

        # Store for use in plot_overview (offer possibility for method chaining)
        self.sample = X.iloc[[idx]]
        self.sample_iloc = idx
        self.tuned_method = tuned_method
        self.surrogate_pred_str = surrogate_pred_str

        return self  # Return self for method chaining

    def plot_overview(
        self, show=True, plot_max_depth=None, colormap=None, plot_gui=False, temp_gui_dir=None
    ):

        if self.sample is None or self.tuned_method is None:
            raise ValueError(
                "Call 'explain()' method first, to generate the explanation and set up the sample."
            )
        sample = self.sample
        tuned_method = self.tuned_method

        if self.verbose >= 0:
            y_pred_orig = predict_helper(self.clf, sample)
            print("Bellatrex prediction:", self.surrogate_pred_str)
            print("Black box prediction: " + frmt_pretty_print(y_pred_orig, digits_single=4))
            print("#" * 58, flush=True)

        if self.verbose >= 4.0:  # print more details in the console:
            if tuned_method.final_trees_idx and tuned_method.cluster_sizes:
                for tree_idx, cluster_size in zip(
                    tuned_method.final_trees_idx, tuned_method.cluster_sizes
                ):
                    rule_print_inline(
                        self.clf[tree_idx],
                        sample,
                        weight=cluster_size / np.sum(tuned_method.cluster_sizes),
                        max_features_print=self.MAX_FEATURE_PRINT,
                    )

        # Set interactive mode to False
        plt.ioff()
        fig, axes = None, None

        # Prepare data and info to be plotted
        plot_kmeans, plot_data_bunch = tuned_method.preselect_represent_cluster_trees()

        if not plot_gui:  # Plot standard overview plots, without interactive features
            if plot_max_depth is not None:
                warnings.warn(
                    f"Max depth for tree visualization = {plot_max_depth} "
                    f"has no effect if plot_gui is set to {plot_gui}"
                )

            fig, axes = plot_preselected_trees(
                plot_data_bunch,
                plot_kmeans,
                tuned_method,
                base_font_size=self.FONT_SIZE,
                colormap=colormap,
            )
            fig.suptitle("Plot overview", fontsize=16)

        else:  # Interactive GUI plotting
            if isinstance(self.clf, EnsembleWrapper):
                raise ValueError(
                    "GUI interface is not compatible with packed EnsembleWrapper yet."
                    "\nPlease use the original sklearn.ensemble class and do not call"
                    "the pack_trained_ensemble function on it."
                )

            matplotlib.use("Agg")
            print("Matplotlib set in a non-interactive backend, with: \"matplotlib.use('Agg')\"")
            from .gui_utils import check_and_import_gui_dependencies

            dearpygui, dearpygui_ext = check_and_import_gui_dependencies()
            from .gui_plots_code import plot_with_interface

            if not show:
                warnings.warn(
                    "Plots are shown immediately while in an interactive session (plot_gui = True)."
                    "\nShow = False is therefore ignored."
                )

            # A temporary directory is used to store, read and clear files created during
            #  the User interactions, a writable GUI temp dir (no __file__, no site-packages)
            base_dir = self._pick_base_dir(out_dir_name="temp-plots")
            temp_files_dir = (
                temp_gui_dir
                if (isinstance(temp_gui_dir, str) and temp_gui_dir.strip())
                else base_dir
            )
            os.makedirs(temp_files_dir, exist_ok=True)

            # Call plot_with_interface and retrieve the interactive plots
            plots = plot_with_interface(
                plot_data_bunch,
                plot_kmeans,
                tuned_method,
                temp_files_dir=temp_files_dir,
                max_depth=plot_max_depth,
                colormap=colormap,
            )

            # Create a placeholder figure and axes to match the expected output format
            fig = plt.figure(figsize=(8, 6))
            axes = plots  # Use the returned plots as axes-like objects

        return fig, axes  # plt.gcf()

    def _pick_base_dir(self, out_dir_name="explanations-output"):
        """
        Choose a base dir that callers can override via env.
        """
        env_dir = os.getenv("BELLATREX_EXPLAIN_DIR")
        # if env_dir:
        #     return env_dir
        # default to cwd/explanations-output
        return os.path.join(os.getcwd(), out_dir_name)

    def _pick_runtime_dir(self, out_dir_name="temp-plots"):
        """
        Choose a runtime dir that callers can override via env.
        """
        env_dir = os.getenv("BELLATREX_RUNTIME_DIR")
        if env_dir:
            return env_dir
        # default to cwd/temp-plots
        return os.path.join(os.getcwd(), out_dir_name)

    def create_rules_txt(self, out_dir="explanations-output", out_file=None):
        """
        Create rules with Bellatrex and stores them as txt files.

        Output path management:
        - 'out_file' is the filename. If None, defaults to 'Btrex_sample_<iloc>.txt'.
        - 'out_dir' controls location:
            * absolute path: used as-is
            * relative path / "." / "": resolved under base = BELLATREX_EXPLAIN_DIR or <cwd>/explanations-output
        - Never writes near site-packages, no tempfile fallback: ensure Docker/CI sets a writable workdir or env.
        """
        tuned_method = self.tuned_method

        # Ensure DataFrame with stable column names
        if isinstance(self.sample, np.ndarray):
            self.sample = pd.DataFrame(self.sample)
            self.sample.columns = [f"X_{i}" for i in range(len(self.sample.columns))]

        # 1) Resolve target_dir from out_dir (absolute respected; relative goes under base)
        if os.path.isabs(out_dir):
            target_dir = out_dir
        else:
            base = self._pick_base_dir(out_dir_name="explanations-output")
            target_dir = base if out_dir in ("", ".") else os.path.join(base, out_dir)

        # filename comes ONLY from out_file or default; strip any accidental path parts
        filename = out_file or f"Btrex_sample_{self.sample_iloc}.txt"
        filename = os.path.basename(filename)  # enforce "filename only"

        main_path = os.path.join(target_dir, filename)
        stem, _ = os.path.splitext(main_path)
        extra_path = f"{stem}_extra.txt"

        os.makedirs(os.path.dirname(main_path), exist_ok=True)

        # 5) Write main file
        with open(main_path, "w", encoding="utf8"):
            pass
        with open(main_path, "a", encoding="utf8") as f:
            for idx, clus_size in zip(tuned_method.final_trees_idx, tuned_method.cluster_sizes):
                rule_to_file(
                    self.clf[idx],
                    self.sample,
                    clus_size / np.sum(tuned_method.cluster_sizes),
                    self.MAX_FEATURE_PRINT,
                    f,
                )
            f.write(f"Bellatrex prediction: {self.surrogate_pred_str}")

        # 6) Write extra file
        with open(extra_path, "w", encoding="utf8"):
            pass
        with open(extra_path, "a", encoding="utf8") as f:
            for idx in range(self.clf.n_estimators):
                if idx not in tuned_method.final_trees_idx:
                    rule_to_file(self.clf[idx], tuned_method.sample, 0, self.MAX_FEATURE_PRINT, f)

        # 7) Parse & validate
        rules, preds, baselines, weights, _ = read_rules(file=main_path, file_extra=extra_path)
        if isinstance(preds[0], list):
            preds = [list(map(float, pred)) for pred in preds]
        if isinstance(baselines[0], list):
            baselines = [list(map(float, baseline)) for baseline in baselines]
        _input_validation(rules, preds, baselines, weights)

        return main_path, extra_path

    def plot_visuals(
        self,
        plot_max_depth=None,
        preds_distr=None,
        conf_level=None,
        tot_digits=4,
        b_box_pred=None,
        keep_files=False,
        out_file=None,
        show=True,
    ):
        """
        Generate a visual explanation of the rules for a single-output task.

        This method creates rule files, parses them, and plots a visual representation
        of the rules extracted by Bellatrex. Intended only for single-output
        setups (e.g., binary classification, regression or survival analysis). For multi-output tasks,
        use `plot_overview()` instead.

        Parameters:
            plot_max_depth (int, optional): Maximum number of rules (depth) to display.
            preds_distr (np.ndarray or Series, optional): Distribution of training predictions,
                used for background histograms or contextual visuals.
            conf_level (float, optional): Confidence level to display for predictions.
            tot_digits (int, default=4): Number of digits to round numeric values in the plot.
            b_box_pred (np.ndarray or Series, optional): Optional black-box prediction reference.
            keep_files (bool, default=False): Whether to keep intermediate `.txt` rule files.
            out_file (str, optional): Custom file path to store generated rule files.
            show (bool, default=True): whether to immediately display the figure.

        Returns:
            Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
                The figure and axes containing the plotted rule visualization, regardless of `show`.

        Raises:
            ValueError: If called on a multi-output setup, which is not supported by this function.
        """
        tuned_method = self.tuned_method
        multi_output_cases = tuned_method.MSA_KEYS + tuned_method.MTC_KEYS + tuned_method.MTR_KEYS

        if tuned_method.set_up in multi_output_cases:
            raise ValueError(
                f"plot_visuals() is compatible with single-output tasks only,\n"
                f"found '{tuned_method.set_up}'. Use plot_overview() instead"
            )

        out_file, file_extra = self.create_rules_txt()
        rules, preds, baselines, weights, other_preds = read_rules(
            file=out_file, file_extra=file_extra
        )

        _input_validation(rules, preds, baselines, weights)

        if keep_files and self.verbose >= 3:
            print(f"Stored Bellatrex rules in:\n{out_file}")

        if keep_files is False:  # delete txt files after reading stored rules
            os.remove(out_file)
            os.remove(file_extra)
            if self.verbose >= 3:
                print(f"Removed txt files from:\n{os.path.dirname(out_file)}")

        fig, axs = plot_rules(
            rules,
            preds,
            baselines,
            weights,
            max_rulelen=plot_max_depth,
            other_preds=other_preds,
            base_fontsize=self.FONT_SIZE - 1,
            conf_level=conf_level,
            tot_digits=tot_digits,
            # cmap='shap',
            preds_distr=preds_distr,
            b_box_pred=b_box_pred,
        )  # b_box_pred=y_test.iloc[idx]

        if show is True:
            plt.show()

        return fig, axs

    def predict_survival_curve(self, X, idx):
        """
        To be implemented
        """
        if not self.set_up in ["surv", "survival"]:
            raise ValueError("Input set-up is not a time-to-event!")
        else:
            raise NotImplementedError("predict_survival_curve is not implemented yet.")

    def predict_median_surv_time(self, X, idx):
        """
        To be implemented
        """
        if not self.set_up in ["surv", "survival"]:
            raise ValueError("Input set-up is not a time-to-event!")
        else:
            raise NotImplementedError("predict_median_surv_time is not implemented yet.")



# -*- coding: utf-8 -*-
"""
Created on Mon Jul 22 16:19:54 2024
bellatrex package initializer.
@author: Klest Dedja
"""

# from typing import TYPE_CHECKING  # so that pylance checker does not complain

try:
    from .__version__ import __version__
except ImportError:
    __version__ = "unknown"


# Lazy import to safely expose BellatrexExplain to
# outer layer only after (ediatble) installation is complete
def __getattr__(name):
    if name == "BellatrexExplain":
        from .bellatrex_explain import BellatrexExplain  # pylint: disable=import-outside-toplevel

        return BellatrexExplain
    raise AttributeError(f"module {__name__} has no attribute {name}")


__all__ = ["BellatrexExplain"]



import pytest
import numpy as np
import matplotlib
matplotlib.use('Agg')  # for headless testing
import matplotlib.pyplot as plt
from bellatrex.visualization import plot_rules, parse, read_rules

# --- plot_rules ---
def test_plot_rules_basic():
    rules = [["feature_1 <= 0.5", "feature_2 > 1.0"]]
    preds = [[0.2, 0.8]]
    baselines = [[0.1]]
    weights = [1.0]
    fig, axs = plot_rules(rules, preds, baselines, weights)
    assert fig is not None
    assert axs is not None
    plt.close(fig)

def test_plot_rules_with_other_preds():
    rules = [["feature_1 <= 0.5", "feature_2 > 1.0"]]
    preds = [[0.2, 0.8]]
    baselines = [[0.1]]
    weights = [1.0]
    other_preds = [[0.15, 0.85]]
    fig, axs = plot_rules(rules, preds, baselines, weights, other_preds=other_preds)
    assert fig is not None
    plt.close(fig)

def test_plot_rules_with_preds_distr():
    rules = [["feature_1 <= 0.5", "feature_2 > 1.0"]]
    preds = [[0.2, 0.8]]
    baselines = [[0.1]]
    weights = [1.0]
    preds_distr = np.random.normal(0.5, 0.1, 100)
    fig, axs = plot_rules(rules, preds, baselines, weights, preds_distr=preds_distr)
    assert fig is not None
    plt.close(fig)

def test_plot_rules_with_conf_level():
    rules = [["feature_1 <= 0.5", "feature_2 > 1.0"]]
    preds = [[0.2, 0.8]]
    baselines = [[0.1]]
    weights = [1.0]
    other_preds = [[0.15, 0.85]]
    fig, axs = plot_rules(rules, preds, baselines, weights, other_preds=other_preds, conf_level=0.9)
    assert fig is not None
    plt.close(fig)

def test_plot_rules_with_bbox_pred():
    rules = [["feature_1 <= 0.5", "feature_2 > 1.0"]]
    preds = [[0.2, 0.8]]
    baselines = [[0.1]]
    weights = [1.0]
    fig, axs = plot_rules(rules, preds, baselines, weights, b_box_pred=0.7)
    assert fig is not None
    plt.close(fig)

# --- parse ---
def test_parse_latex_and_parenthesis():
    s = "feature_1 <= 0.5 (val=0.3)"
    out = parse(s)
    assert r"$\\leq$" in out or "$\\leq$" in out
    s2 = "feature_2 >= 1.0 (val=1.2)"
    out2 = parse(s2)
    assert r"$\\geq$" in out2 or "$\\geq$" in out2

def test_parse_removes_after_paren():
    s = "feature_1 <= 0.5 (val=0.3)"
    out = parse(s)
    assert "(" not in out or out.endswith("(")

# --- read_rules ---
def test_read_rules(tmp_path):
    rule_file = tmp_path / "rules.txt"
    extra_file = tmp_path / "rules_extra.txt"
    rule_file.write_text(
        """
RULE WEIGHT: 0.5
Baseline prediction: 0.1
node: feature_1 <= 0.5 --> 0.2
node: feature_2 > 1.0 --> 0.8
leaf
Bellatrex prediction: 0.8
"""
    )
    extra_file.write_text(
        """
Baseline prediction: 0.1
node: feature_1 <= 0.5 --> 0.2
node: feature_2 > 1.0 --> 0.8
leaf
"""
    )
    rules, preds, baselines, weights, other_preds = read_rules(str(rule_file), str(extra_file))
    assert isinstance(rules, list)
    assert isinstance(preds, list)
    assert isinstance(baselines, list)
    assert isinstance(weights, list)
    assert isinstance(other_preds, list)
    # Test without extra file
    rules2, preds2, baselines2, weights2, other_preds2 = read_rules(str(rule_file))
    assert other_preds2 is None



def check_and_import_gui_dependencies():
    try:
        import dearpygui
        import dearpygui_ext
    except ImportError as e:
        raise ImportError(
            "Optional dependencies for the GUI are not installed, "
            "Namely dearpygui>=1.6.2 and dearpygui-ext>=0.9.5. "
            "Please install them using: pip install bellatrex[gui]"
        ) from e
    return dearpygui, dearpygui_ext



import pytest
import numpy as np
from scipy.sparse import csr_matrix
from bellatrex.wrapper_class import (
    pack_trained_ensemble,
    EnsembleWrapper,
    tree_to_dict,
    tree_list_to_model,
)
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sksurv.ensemble import RandomSurvivalForest
from sksurv.tree import SurvivalTree

# --- Helper for mock tree dicts ---
def make_tree_dict(learner_class="DecisionTreeClassifier", n_outputs=1):
    # Minimal tree dict for EnsembleWrapper
    return {
        "features": [0, 1, -1],
        "n_leaves": 2,
        "node_sample_weight": [10, 5, 5],
        "children_left": [1, -1, -1],
        "children_right": [2, -1, -1],
        "thresholds": [0.5, -2, -2],
        "values": np.array([[0.4, 0.6], [0.6, 0.4], [0.2, 0.8]]) if n_outputs == 2 else np.array([[0.6], [0.4], [0.8]]),
        "base_offset": 0.6,
        "max_depth": 1,
        "feature_names_in_": ["f1", "f2"],
        "n_features_in_": 2,
        "learner_class": learner_class,
        "output_format": "auto",
        "ensemble_class": "RandomForestClassifier"
    }

def test_pack_trained_ensemble_classifier():
    clf = RandomForestClassifier(n_estimators=2, random_state=0).fit(np.random.rand(10, 2), np.random.randint(0, 2, 10))
    packed = pack_trained_ensemble(clf, set_up="auto")
    assert isinstance(packed, dict)
    assert "trees" in packed
    assert packed["ensemble_class"] == "RandomForestClassifier"

def test_pack_trained_ensemble_regressor():
    clf = RandomForestRegressor(n_estimators=2, random_state=0).fit(np.random.rand(10, 2), np.random.rand(10))
    packed = pack_trained_ensemble(clf, set_up="auto")
    assert isinstance(packed, dict)
    assert "trees" in packed
    assert packed["ensemble_class"] == "RandomForestRegressor"

def test_pack_trained_ensemble_invalid():
    with pytest.raises(ValueError):
        pack_trained_ensemble(object(), set_up="auto")

def test_ensemble_wrapper_predict_and_decision_path():
    tree_dict = make_tree_dict()
    model_dict = tree_list_to_model([tree_dict, tree_dict])
    wrapper = EnsembleWrapper(model_dict)
    X = np.random.rand(5, 2)
    preds = wrapper.predict(X)
    assert preds.shape == (5, wrapper.n_outputs_)
    paths, ptr = wrapper.decision_path(X)
    assert isinstance(paths, csr_matrix)
    assert isinstance(ptr, np.ndarray)
    # __getitem__
    est = wrapper[0]
    assert hasattr(est, "predict")

def test_estimator_predict_and_decision_path():
    tree_dict = make_tree_dict()
    est = EnsembleWrapper.Estimator(tree_dict)
    X = np.random.rand(3, 2)
    preds = est.predict(X)
    assert preds.shape[0] == 3
    path = est.decision_path(X)
    assert isinstance(path, csr_matrix)

def test_tree_to_dict_binary():
    clf = RandomForestClassifier(n_estimators=1, random_state=0).fit(np.random.rand(10, 2), np.random.randint(0, 2, 10))
    d = tree_to_dict(clf, 0, output_format="auto")
    assert isinstance(d, dict)
    assert "values" in d

def test_tree_to_dict_regression():
    clf = RandomForestRegressor(n_estimators=1, random_state=0).fit(np.random.rand(10, 2), np.random.rand(10))
    d = tree_to_dict(clf, 0, output_format="auto")
    assert isinstance(d, dict)
    assert "values" in d

def test_tree_to_dict_invalid():
    class Dummy:
        pass
    with pytest.raises(Exception):
        tree_to_dict(Dummy(), 0, output_format="auto")

def test_tree_list_to_model_consistency():
    tree_dict1 = make_tree_dict()
    tree_dict2 = make_tree_dict()
    tree_dict1["output_format"] = tree_dict2["output_format"] = "auto"
    tree_dict1["ensemble_class"] = tree_dict2["ensemble_class"] = "RandomForestClassifier"
    model = tree_list_to_model([tree_dict1, tree_dict2])
    assert isinstance(model, dict)
    assert model["ensemble_class"] == "RandomForestClassifier"
    assert model["output_format"] == "auto"
    assert "trees" in model

def test_tree_to_dict_error_branches():
    # Test error for unrecognized combination
    class DummyTree:
        n_outputs_ = 99
        def __init__(self):
            self.value = np.zeros((3, 3, 3))
            self.children_left = [1, -1, -1]
            self.children_right = [2, -1, -1]
            self.n_leaves = 2
            self.max_depth = 1
            self.feature = [0, 1, -1]
            self.threshold = [0.5, -2, -2]
            self.weighted_n_node_samples = [10, 5, 5]
            self.tree_ = self
    class DummyClf:
        def __getitem__(self, idx):
            return DummyTree()
        __class__ = type("Dummy", (), {})
    with pytest.raises(ValueError):
        tree_to_dict(DummyClf(), 0, output_format="unknown")



import numpy as np
import pandas as pd
import pytest
from sklearn.datasets import make_classification
from sklearn.ensemble import RandomForestClassifier

from bellatrex.tree_dissimilarity import TreeDissimilarity
from bellatrex.tree_representation_utils import rule_splits_to_vector


def test_avg_dissimilarity_computation():

    vector_1 = np.array([0, 0.2, 0.8, 0])
    vector_2 = np.array([0.3, 0, 0.7, 0])

    expected_dissimilarity = 0.461538461  # ( equal to 1 - 0.7/1.3)

    dissim_class = TreeDissimilarity(
        clf=None, set_up=None, tree_indeces=None
    )  # TODO: Vectorization functions do not need to be class methods

    dissim_matrix = dissim_class.vectors_to_dissim_matrix([vector_1, vector_2])
    avg_dissimilarity = dissim_class.compute_avg_dissimilarity(dissim_matrix)

    assert np.isclose(avg_dissimilarity, expected_dissimilarity, rtol=1e-5)


def test_dissimilarity_matrix_computation():
    vector_1 = np.array([0, 0.2, 0.8, 0])
    vector_2 = np.array([0.3, 0, 0.7, 0])
    vector_3 = np.array([0.5, 0.5, 0, 0])

    # Manually computed dissimilarity matrix elements:
    expected_matrix = np.array(
        [
            [0.0, 0.461538461, 0.888888888],
            [0.461538461, 0.0, 0.82352941],
            [0.888888888, 0.82352941, 0.0],
        ]
    )
    dissim_object = TreeDissimilarity(clf=None, set_up=None, tree_indeces=None)
    dissim_matrix = dissim_object.vectors_to_dissim_matrix([vector_1, vector_2, vector_3])

    assert np.allclose(dissim_matrix, expected_matrix, atol=1e-5)


def test_avg_dissim_and_matrix_function():

    # Generate deterministic data
    X, y = make_classification(n_samples=30, n_features=5, random_state=42)

    # Fit deterministic forest
    clf = RandomForestClassifier(n_estimators=2, max_depth=3, random_state=0)
    clf.fit(X, y)

    test_sample = pd.Series(X[0])
    dissim_object = TreeDissimilarity(
        clf=clf, set_up="binary", tree_indeces=[0, 1], sample=test_sample
    )  # TODO: Vectorization functions do not need to be class methods

    vector_1 = rule_splits_to_vector(clf, 0, feature_represent="weighted", sample=test_sample)
    vector_2 = rule_splits_to_vector(clf, 1, feature_represent="weighted", sample=test_sample)

    dissim_matrix = dissim_object.vectors_to_dissim_matrix([vector_1, vector_2])
    avg_dissimilarity = dissim_object.compute_avg_dissimilarity(dissim_matrix)

    alternative_avg_dissim, alternative_matrix = dissim_object.return_avg_dissim_and_matrix()

    assert np.isclose(avg_dissimilarity, alternative_avg_dissim, rtol=1e-5)
    assert np.allclose(dissim_matrix, alternative_matrix, rtol=1e-5)


# print("Let's get started")
test_avg_dissim_and_matrix_function()
# # test_real_tree_vectors()
print("Test completed successfully.")



from importlib.resources import files
import pandas as pd
import numpy as np


def load_data(filename: str) -> pd.DataFrame:
    filepath = files("bellatrex.datasets") / filename
    return pd.read_csv(filepath)


# Functions to load specific datasets
# Separation between X and y is done 'manually', that is: case by case
def load_binary_data(return_X_y=False):
    if return_X_y is False:
        return load_data("bin_tutorial.csv")
    else:
        df = load_data("bin_tutorial.csv")
        X = df.iloc[:, :-1]
        y = df.iloc[:, -1]
        return X, y


def load_regression_data(return_X_y=False):
    if return_X_y is False:
        return load_data("regress_tutorial.csv")
    else:
        df = load_data("regress_tutorial.csv")
        X = df.iloc[:, :-1]
        y = df.iloc[:, -1]
        return X, y


def load_survival_data(return_X_y=False):
    if return_X_y is False:
        return load_data("surv_tutorial.csv")
    else:
        df = load_data("surv_tutorial.csv")
        X = df.iloc[:, :-2]
        y = df.iloc[:, -2:]

        dtypes_map = {str(y.columns[0]): np.bool_, str(y.columns[1]): np.float32}
        y = y.to_records(index=False, column_dtypes=dtypes_map)

        return X, y


def load_mlc_data(return_X_y=False):
    if return_X_y is False:
        return load_data("mlc_tutorial.csv")
    else:
        df = load_data("mlc_tutorial.csv")
        columns_out = [col for col in df.columns if "tag" in col]
        columns_in = [col for col in df.columns if "tag" not in col]

        return df.loc[:, columns_in], df.loc[:, columns_out]


def load_mtr_data(return_X_y=False):
    if return_X_y is False:
        return load_data("mtr_tutorial.csv")
    else:
        df = load_data("mtr_tutorial.csv")
        columns_out = [col for col in df.columns if "target" in col]
        columns_in = [col for col in df.columns if "target" not in col]

        return df.loc[:, columns_in], df.loc[:, columns_out]
